from typing import Optional, Tuple, Union

import numpy as np

from probnum import diffeq, filtsmooth, problems, randvars, statespace
from probnum.type import FloatArgType, IntArgType

__all__ = [
    "benes_daum",
    "car_tracking",
    "logistic_ode",
    "ornstein_uhlenbeck",
    "pendulum",
]


def car_tracking(
    measurement_variance: FloatArgType = 0.5,
    process_diffusion: FloatArgType = 1.0,
    model_ordint: IntArgType = 1,
    timespan: Tuple[FloatArgType, FloatArgType] = (0.0, 20.0),
    step: FloatArgType = 0.2,
    initrv: Optional[randvars.RandomVariable] = None,
    forward_implementation: str = "classic",
    backward_implementation: str = "classic",
):
    r"""Filtering/smoothing setup for a simple car-tracking scenario.

    A discrete, linear, time-invariant Gaussian state space model for car-tracking,
    based on Example 3.6 in Särkkä, 2013. [1]_
    Let :math:`X = (\dot{x}_1, \dot{x}_2, \ddot{x}_1, \ddot{x}_2)`. Then the state
    space model has the following discretized formulation

    .. math::

        X(t_{n}) &=
        \begin{pmatrix}
          1 & 0 & \Delta t& 0 \\
          0 & 1 & 0 & \Delta t \\
          0 & 0 & 1 & 0 \\
          0 & 0 & 0 & 1
        \end{pmatrix} X(t_{n-1})
        +
        q_n \\
        y_{n} &=
        \begin{pmatrix}
          1 & 0 & 0 & 0 \\
          0 & 1 & 0 & 0 \\
        \end{pmatrix} X(t_{n})
        + r_n

    where :math:`q_n \sim \mathcal{N}(0, Q)` and :math:`r_n \sim \mathcal{N}(0, R)`
    for process noise covariance matrix :math:`Q` and measurement noise covariance
    matrix :math:`R`.

    Parameters
    ----------
    measurement_variance
        Marginal measurement variance.
    process_diffusion
        Diffusion constant for the dynamics.
    model_ordint
        Order of integration for the dynamics model. Defaults to one, which corresponds
        to a Wiener velocity model.
    timespan
        :math:`t_0` and :math:`t_{\max}` of the time grid.
    step
        Step size of the time grid.
    initrv
        Initial random variable.

    Returns
    -------
    regression_problem
        ``RegressionProblem`` object with time points and noisy observations.
    statespace_components
        Dictionary containing

        * dynamics model
        * measurement model
        * initial random variable

    References
    ----------
    .. [1] Särkkä, Simo. Bayesian Filtering and Smoothing. Cambridge University Press,
        2013.

    """
    state_dim = 2
    model_dim = state_dim * (model_ordint + 1)
    measurement_dim = 2
    dynamics_model = statespace.IBM(
        ordint=model_ordint,
        spatialdim=state_dim,
        forward_implementation=forward_implementation,
        backward_implementation=backward_implementation,
    )
    dynamics_model.dispmat *= process_diffusion

    discrete_dynamics_model = dynamics_model.discretise(dt=step)

    measurement_matrix = np.eye(measurement_dim, model_dim)
    measurement_cov = measurement_variance * np.eye(measurement_dim)
    measurement_cov_cholesky = np.sqrt(measurement_variance) * np.eye(measurement_dim)
    measurement_model = statespace.DiscreteLTIGaussian(
        state_trans_mat=measurement_matrix,
        shift_vec=np.zeros(measurement_dim),
        proc_noise_cov_mat=measurement_cov,
        proc_noise_cov_cholesky=measurement_cov_cholesky,
        forward_implementation=forward_implementation,
        backward_implementation=backward_implementation,
    )

    if initrv is None:
        initrv = randvars.Normal(
            np.zeros(model_dim),
            measurement_variance * np.eye(model_dim),
            cov_cholesky=np.sqrt(measurement_variance) * np.eye(model_dim),
        )

    # Set up regression problem
    time_grid = np.arange(*timespan, step=step)
    states, obs = statespace.generate_samples(
        dynmod=discrete_dynamics_model,
        measmod=measurement_model,
        initrv=initrv,
        times=time_grid,
    )
    regression_problem = problems.RegressionProblem(
        observations=obs, locations=time_grid, solution=states
    )

    statespace_components = dict(
        dynamics_model=discrete_dynamics_model,
        measurement_model=measurement_model,
        initrv=initrv,
    )
    return regression_problem, statespace_components


def ornstein_uhlenbeck(
    measurement_variance: FloatArgType = 0.1,
    driftspeed: FloatArgType = 0.21,
    process_diffusion: FloatArgType = 0.5,
    time_grid: Optional[np.ndarray] = None,
    initrv: Optional[randvars.RandomVariable] = None,
    forward_implementation: str = "classic",
    backward_implementation: str = "classic",
):
    r"""Filtering/smoothing setup based on an Ornstein Uhlenbeck process.

    A linear, time-invariant state space model for the dynamics of a time-invariant
    Ornstein-Uhlenbeck process. See e.g. Example 10.19 in Särkkä et. al, 2019. [1]_
    Here, we formulate a continuous-discrete state space model:

    .. math::

        d x(t) &= \lambda x(t) d t + L d w(t) \\
        y_n &= x(t_n) + r_n

    for a drift constant :math:`\lambda` and a driving Wiener process :math:`w(t)`.
    :math:`r_n \sim \mathcal{N}(0, R)` is Gaussian distributed measurement noise
    with covariance matrix :math:`R`.
    Note that the linear, time-invariant dynamics have an equivalent discretization.

    Parameters
    ----------
    measurement_variance
        Marginal measurement variance.
    driftspeed
        Drift parameter of the Ornstein-Uhlenbeck process.
    process_diffusion
        Diffusion constant for the dynamics
    time_grid
        Time grid for the filtering/smoothing problem.
    initrv
        Initial random variable.


    Returns
    -------
    regression_problem
        ``RegressionProblem`` object with time points and noisy observations.
    statespace_components
        Dictionary containing

        * dynamics model
        * measurement model
        * initial random variable


    References
    ----------
    .. [1] Särkkä, Simo, and Solin, Arno. Applied Stochastic Differential Equations.
        Cambridge University Press, 2019
    """

    dynamics_model = statespace.IOUP(
        ordint=0,
        spatialdim=1,
        driftspeed=driftspeed,
        forward_implementation=forward_implementation,
        backward_implementation=backward_implementation,
    )
    dynamics_model.dispmat *= process_diffusion

    measurement_model = statespace.DiscreteLTIGaussian(
        state_trans_mat=np.eye(1),
        shift_vec=np.zeros(1),
        proc_noise_cov_mat=measurement_variance * np.eye(1),
        forward_implementation=forward_implementation,
        backward_implementation=backward_implementation,
    )

    if initrv is None:
        initrv = randvars.Normal(10.0 * np.ones(1), np.eye(1))

    # Set up regression problem
    if time_grid is None:
        time_grid = np.arange(0.0, 20.0, step=0.2)
    states, obs = statespace.generate_samples(
        dynmod=dynamics_model, measmod=measurement_model, initrv=initrv, times=time_grid
    )
    regression_problem = problems.RegressionProblem(
        observations=obs, locations=time_grid, solution=states
    )

    statespace_components = dict(
        dynamics_model=dynamics_model,
        measurement_model=measurement_model,
        initrv=initrv,
    )
    return regression_problem, statespace_components


def pendulum(
    measurement_variance: FloatArgType = 0.1024,
    timespan: Tuple[FloatArgType, FloatArgType] = (0.0, 4.0),
    step: FloatArgType = 0.0075,
    initrv: Optional[randvars.RandomVariable] = None,
):
    r"""Filtering/smoothing setup for a (noisy) pendulum.

    A non-linear, discretized state space model for a pendulum with unknown forces
    acting on the dynamics, modeled as Gaussian noise.
    See e.g. Särkkä, 2013 [1]_ for more details.

    .. math::

        \begin{pmatrix}
          x_1(t_n) \\
          x_2(t_n)
        \end{pmatrix}
        &=
        \begin{pmatrix}
          x_1(t_{n-1}) + x_2(t_{n-1}) \cdot h \\
          x_2(t_{n-1}) - g \sin(x_1(t_{n-1})) \cdot h
        \end{pmatrix}
        +
        q_n \\
        y_n &\sim \sin(x_1(t_n)) + r_n

    for some ``step`` size :math:`h` and Gaussian process noise
    :math:`q_n \sim \mathcal{N}(0, Q)` with

    .. math::
        Q =
        \begin{pmatrix}
          \frac{h^3}{3} & \frac{h^2}{2} \\
          \frac{h^2}{2} & h
        \end{pmatrix}

    :math:`g` denotes the gravitational constant and :math:`r_n \sim \mathcal{N}(0, R)`
    is Gaussian mesurement noise with some covariance :math:`R`.

    Parameters
    ----------
    measurement_variance
        Marginal measurement variance.
    timespan
        :math:`t_0` and :math:`t_{\max}` of the time grid.
    step
        Step size of the time grid.
    initrv
        Initial random variable.

    Returns
    -------
    regression_problem
        ``RegressionProblem`` object with time points and noisy observations.
    statespace_components
        Dictionary containing

        * dynamics model
        * measurement model
        * initial random variable


    References
    ----------
    .. [1] Särkkä, Simo. Bayesian Filtering and Smoothing. Cambridge University Press,
        2013.

    """

    # Graviational constant
    g = 9.81

    # Define non-linear dynamics and measurements
    def f(t, x):
        x1, x2 = x
        y1 = x1 + x2 * step
        y2 = x2 - g * np.sin(x1) * step
        return np.array([y1, y2])

    def df(t, x):
        x1, x2 = x
        y1 = [1, step]
        y2 = [-g * np.cos(x1) * step, 1]
        return np.array([y1, y2])

    def h(t, x):
        x1, x2 = x
        return np.array([np.sin(x1)])

    def dh(t, x):
        x1, x2 = x
        return np.array([[np.cos(x1), 0.0]])

    process_noise_cov = (
        np.diag(np.array([step ** 3 / 3, step]))
        + np.diag(np.array([step ** 2 / 2]), 1)
        + np.diag(np.array([step ** 2 / 2]), -1)
    )

    dynamics_model = statespace.DiscreteGaussian(
        input_dim=2,
        output_dim=2,
        state_trans_fun=f,
        proc_noise_cov_mat_fun=lambda t: process_noise_cov,
        jacob_state_trans_fun=df,
    )

    measurement_model = statespace.DiscreteGaussian(
        input_dim=2,
        output_dim=1,
        state_trans_fun=h,
        proc_noise_cov_mat_fun=lambda t: measurement_variance * np.eye(1),
        jacob_state_trans_fun=dh,
    )

    if initrv is None:
        initrv = randvars.Normal(np.ones(2), measurement_variance * np.eye(2))

    # Generate data
    time_grid = np.arange(*timespan, step=step)
    states, obs = statespace.generate_samples(
        dynmod=dynamics_model, measmod=measurement_model, initrv=initrv, times=time_grid
    )
    regression_problem = problems.RegressionProblem(
        observations=obs, locations=time_grid, solution=states
    )
    statespace_components = dict(
        dynamics_model=dynamics_model,
        measurement_model=measurement_model,
        initrv=initrv,
    )
    return regression_problem, statespace_components


def benes_daum(
    measurement_variance: FloatArgType = 0.1,
    process_diffusion: FloatArgType = 1.0,
    time_grid: Optional[np.ndarray] = None,
    initrv: Optional[randvars.RandomVariable] = None,
):
    r"""Filtering/smoothing setup based on the Beneš SDE.

    A non-linear state space model for the dynamics of a Beneš SDE.
    Here, we formulate a continuous-discrete state space model:

    .. math::

        d x(t) &= \tanh(x(t)) d t + L d w(t) \\
        y_n &= x(t_n) + r_n

    for a driving Wiener process :math:`w(t)` and Gaussian distributed measurement noise
    :math:`r_n \sim \mathcal{N}(0, R)` with measurement noise
    covariance matrix :math:`R`.

    Parameters
    ----------
    measurement_variance
        Marginal measurement variance.
    process_diffusion
        Diffusion constant for the dynamics
    time_grid
        Time grid for the filtering/smoothing problem.
    initrv
        Initial random variable.

    Returns
    -------
    regression_problem
        ``RegressionProblem`` object with time points and noisy observations.
    statespace_components
        Dictionary containing

        * dynamics model
        * measurement model
        * initial random variable

    Notes
    -----
    In order to generate observations for the returned ``RegressionProblem`` object,
    the non-linear Beneš SDE has to be linearized.
    Here, a ``ContinuousEKFComponent`` is used, which corresponds to a first-order
    linearization as used in the extended Kalman filter.
    """

    def f(t, x):
        return np.tanh(x)

    def df(t, x):
        return 1.0 - np.tanh(x) ** 2

    def l(t):
        return process_diffusion * np.ones(1)

    if initrv is None:
        initrv = randvars.Normal(np.zeros(1), 3.0 * np.eye(1))

    dynamics_model = statespace.SDE(dimension=1, driftfun=f, dispmatfun=l, jacobfun=df)
    measurement_model = statespace.DiscreteLTIGaussian(
        state_trans_mat=np.eye(1),
        shift_vec=np.zeros(1),
        proc_noise_cov_mat=measurement_variance * np.eye(1),
    )

    # Generate data
    if time_grid is None:
        time_grid = np.arange(0.0, 4.0, step=0.2)
    # The non-linear dynamics are linearized according to an EKF in order
    # to generate samples.
    linearized_dynamics_model = filtsmooth.ContinuousEKFComponent(
        non_linear_model=dynamics_model
    )
    states, obs = statespace.generate_samples(
        dynmod=linearized_dynamics_model,
        measmod=measurement_model,
        initrv=initrv,
        times=time_grid,
    )
    regression_problem = problems.RegressionProblem(
        observations=obs, locations=time_grid, solution=states
    )

    statespace_components = dict(
        dynamics_model=dynamics_model,
        measurement_model=measurement_model,
        initrv=initrv,
    )
    return regression_problem, statespace_components


def logistic_ode(
    y0: Optional[Union[np.ndarray, FloatArgType]] = None,
    timespan: Tuple[FloatArgType, FloatArgType] = (0.0, 2.0),
    step: FloatArgType = 0.1,
    params: Tuple[FloatArgType, FloatArgType] = (6.0, 1.0),
    initrv: Optional[randvars.RandomVariable] = None,
    evlvar: Optional[Union[np.ndarray, FloatArgType]] = None,
    ek0_or_ek1: IntArgType = 1,
    order: IntArgType = 3,
    forward_implementation: str = "classic",
    backward_implementation: str = "classic",
):
    r"""Filtering/smoothing setup for a probabilistic ODE solver for the logistic ODE.

    This state space model assumes an integrated Brownian motion prior on the dynamics
    and constructs the ODE likelihood based on the vector field defining the
    logistic ODE.

    Parameters
    ----------
    y0
        Initial conditions of the Initial Value Problem
    timespan
        Time span of the problem
    params
        Parameters for the logistic ODE
    initrv
        Initial random variable of the probabilistic ODE solver
    evlvar
        See :py:class:`probnum.diffeq.GaussianIVPFilter`
    ek0_or_ek1
        See :py:class:`probnum.diffeq.GaussianIVPFilter`
    order
        Order of integration for the Integrated Brownian Motion prior of the solver.

    Returns
    -------
    regression_problem
        ``RegressionProblem`` object with time points and zero-observations.

    statespace_components
        Dictionary containing

        * dynamics model
        * measurement model
        * initial random variable
        * The initial value problem based on the logistic ODE.

    See Also
    --------
    :py:class:`probnum.diffeq.GaussianIVPFilter`

    """

    if y0 is None:
        y0 = np.array([0.1])
    y0 = np.atleast_1d(y0)

    if evlvar is None:
        evlvar = np.zeros((1, 1))

    logistic_ivp = diffeq.logistic(
        timespan=timespan, initrv=randvars.Constant(y0), params=params
    )
    dynamics_model = statespace.IBM(
        ordint=order,
        spatialdim=1,
        forward_implementation=forward_implementation,
        backward_implementation=backward_implementation,
    )
    measurement_model = filtsmooth.DiscreteEKFComponent.from_ode(
        logistic_ivp,
        prior=dynamics_model,
        evlvar=evlvar,
        ek0_or_ek1=ek0_or_ek1,
        forward_implementation=forward_implementation,
        backward_implementation=backward_implementation,
    )

    if initrv is None:
        initmean = np.array([0.1, 0, 0.0, 0.0])
        initcov = np.diag([0.0, 1.0, 1.0, 1.0])
        initrv = randvars.Normal(initmean, initcov)

    # Generate zero-data
    time_grid = np.arange(*timespan, step=step)
    solution = logistic_ivp.solution(time_grid)
    regression_problem = problems.RegressionProblem(
        observations=np.zeros(shape=(time_grid.size, 1)),
        locations=time_grid,
        solution=solution,
    )

    statespace_components = dict(
        dynamics_model=dynamics_model,
        measurement_model=measurement_model,
        initrv=initrv,
        ivp=logistic_ivp,
    )
    return regression_problem, statespace_components



"""Tests for the categorical random variable."""


import string

import numpy as np
import pytest

from probnum import randvars, utils

NDIM = 5

all_supports = pytest.mark.parametrize(
    "support",
    [
        None,
        np.arange(NDIM),
        np.array(list(string.ascii_lowercase)[:NDIM]),
        np.random.rand(NDIM, 3, 3),
    ],
)

all_random_states = pytest.mark.parametrize(
    "random_state",
    [
        None,
        1,
        np.random.default_rng(),
    ],
)


@pytest.fixture
def probabilities():
    probabilities = np.random.rand(NDIM)
    return probabilities / np.sum(probabilities)


@pytest.fixture
def categ(probabilities, support, random_state):
    return randvars.Categorical(
        probabilities=probabilities, support=support, random_state=random_state
    )


@all_random_states
@all_supports
def test_probabilities(categ, probabilities):
    assert categ.probabilities.shape == (NDIM,)
    np.testing.assert_allclose(categ.probabilities, probabilities)


@all_supports
@all_random_states
def test_support(categ):
    assert len(categ.support) == NDIM
    assert isinstance(categ.support, np.ndarray)


@all_supports
@all_random_states
@pytest.mark.parametrize("size", [(), 1, (1,), (1, 1)])
def test_sample(categ, size):
    samples = categ.sample(size=size)
    expected_shape = utils.as_shape(size) + categ.shape
    assert samples.shape == expected_shape


@all_supports
@all_random_states
@pytest.mark.parametrize("index", [0, -1])
def test_pmf(categ, index):
    pmf_value = categ.pmf(x=categ.support[index])
    np.testing.assert_almost_equal(pmf_value, categ.probabilities[index])


@all_supports
@all_random_states
def test_pmf_zero(categ):
    """Make a new Categorical RV that excludes the final point and check that the pmf
    rightfully evaluates to zero."""

    new_categ = randvars.Categorical(
        support=categ.support[:-1],
        probabilities=categ.probabilities[:-1],
        random_state=categ.random_state,
    )
    zero_pmf_value = new_categ.pmf(x=categ.support[-1])
    np.testing.assert_almost_equal(zero_pmf_value, 0.0)


def test_pmf_valueerror():
    """If a PMF has string-valued support, its pmf cannot be evaluated at an integer.

    This value error is intended to guard against the issue presented in
    https://stackoverflow.com/questions/45020217/numpy-where-function-throws-a-futurewarning-returns-scalar-instead-of-list
    """
    categ = randvars.Categorical(probabilities=[0.5, 0.5], support=["a", "b"])
    with pytest.raises(ValueError):
        categ.pmf(2)


@all_supports
@all_random_states
def test_mode(categ):
    mode = categ.mode
    assert mode.shape == categ.shape


@all_supports
@all_random_states
def test_resample(categ):
    new_categ = categ.resample()

    assert isinstance(new_categ, randvars.Categorical)
    assert new_categ.shape == categ.shape

    # Assert all probabilities are equal
    np.testing.assert_allclose(np.diff(new_categ.probabilities), 0.0)

    # Assert support is chosen from old support
    assert np.all(np.isin(new_categ.support, categ.support))



"""Ordinary differential equation classes."""

from .ivp import IVP
from .ivp_examples import fitzhughnagumo, logistic, lorenz, lotkavolterra, seir
from .ode import ODE

__all__ = [
    "ODE",
    "IVP",
    "logistic",
    "fitzhughnagumo",
    "seir",
    "lotkavolterra",
    "lorenz",
]



from typing import Tuple

import numpy as np
import pytest
import pytest_cases

import probnum as pn


@pytest.mark.parametrize(
    "A,B",
    [
        (np.array([[2, -3.5], [12, 6.5]]), np.eye(3)),
        (np.array([[1, -2], [-2.2, 5]]), np.array([[1, -3], [0, -0.5]])),
        (np.array([[4, 1, 4], [2, 3, 2]]), np.array([[-1, 4], [2, 1]])),
        (np.array([[0.4, 2, 0.8], [-0.4, 0, -0.9]]), np.array([[1, 4]])),
        (np.array([[4, 1], [2, 3]]), np.array([[-1, 4], [2, 1]])),
        (
            np.array([[0.4, 2, 0.8], [-0.4, 0, -0.9], [1, 0, 2]]),
            np.array([[1, 4, 0], [-3, -0.4, -100], [0.18, -2, 10]]),
        ),
        (
            np.array([[0.4, 2, 0.8], [-0.4, 0, -0.9], [1, 0, 2]]),
            np.array([[1, 4, 0], [-3, -0.4, -100]]),
        ),
    ],
)
def case_kronecker(
    A: np.ndarray, B: np.ndarray
) -> Tuple[pn.linops.LinearOperator, np.ndarray]:
    linop = pn.linops.Kronecker(A, B)
    matrix = np.kron(A, B)

    return linop, matrix


@pytest.mark.parametrize(
    "A,B",
    [
        (np.array([[1, -2], [-2.2, 5]]), np.array([[1, -3], [0, -0.5]])),
        (np.array([[4, 1], [2, 3]]), np.array([[-1, 4], [2, 1]])),
        (
            np.array([[0.4, 2, 0.8], [-0.4, 0, -0.9], [1, 0, 2]]),
            np.array([[1, 4, 0], [-3, -0.4, -100], [0.18, -2, 10]]),
        ),
    ],
)
@pytest_cases.case(tags=["symmetric_kronecker"])
def case_symmetric_kronecker(
    A: np.ndarray, B: np.ndarray
) -> Tuple[pn.linops.LinearOperator, np.ndarray]:
    linop = pn.linops.SymmetricKronecker(A, B)
    matrix = (np.kron(A, B) + np.kron(B, A)) / 2

    return linop, matrix


@pytest.mark.parametrize(
    "A",
    [
        np.array([[1, -2], [-2.2, 5]]),
        np.array([[1, -3], [0, -0.5]]),
        np.array([[4, 1], [2, 3]]),
        np.array([[-1, 4], [2, 1]]),
        np.array([[0.4, 2, 0.8], [-0.4, 0, -0.9], [1, 0, 2]]),
        np.array([[1, 4, 0], [-3, -0.4, -100], [0.18, -2, 10]]),
    ],
)
@pytest_cases.case(tags=["symmetric_kronecker"])
def case_symmetric_kronecker_identical_factors(
    A: np.ndarray,
) -> Tuple[pn.linops.LinearOperator, np.ndarray]:
    linop = pn.linops.SymmetricKronecker(A)
    matrix = np.kron(A, A)

    return linop, matrix



"""Exponentiated quadratic kernel."""

from typing import Optional

import numpy as np
import scipy.spatial.distance

import probnum.utils as _utils
from probnum.type import IntArgType, ScalarArgType

from ._kernel import Kernel

_InputType = np.ndarray


class ExpQuad(Kernel[_InputType]):
    """Exponentiated quadratic / RBF kernel.

    Covariance function defined by :math:`k(x_0, x_1) = \\exp \\big(-\\frac{\\lVert
    x_0 - x_1 \\rVert^2}{2l^2}\\big)`. This kernel is also known as the squared
    exponential or radial basis function kernel.

    Parameters
    ----------
    input_dim :
        Input dimension of the kernel.
    lengthscale
        Lengthscale of the kernel. Describes the input scale on which the process
        varies.

    See Also
    --------
    RatQuad : Rational quadratic kernel.
    Matern : Matern kernel.

    Examples
    --------
    >>> import numpy as np
    >>> from probnum.kernels import ExpQuad
    >>> K = ExpQuad(input_dim=1, lengthscale=0.1)
    >>> K(np.linspace(0, 1, 3)[:, None])
    array([[1.00000000e+00, 3.72665317e-06, 1.92874985e-22],
           [3.72665317e-06, 1.00000000e+00, 3.72665317e-06],
           [1.92874985e-22, 3.72665317e-06, 1.00000000e+00]])
    """

    def __init__(self, input_dim: IntArgType, lengthscale: ScalarArgType = 1.0):
        self.lengthscale = _utils.as_numpy_scalar(lengthscale)
        super().__init__(input_dim=input_dim, output_dim=1)

    def __call__(self, x0: _InputType, x1: Optional[_InputType] = None) -> np.ndarray:

        x0, x1, kernshape = self._check_and_reshape_inputs(x0, x1)

        # Compute pairwise euclidean distances ||x0 - x1|| / l
        if x1 is None:
            pdists = scipy.spatial.distance.squareform(
                scipy.spatial.distance.pdist(x0 / self.lengthscale, metric="euclidean")
            )
        else:
            pdists = scipy.spatial.distance.cdist(
                x0 / self.lengthscale, x1 / self.lengthscale, metric="euclidean"
            )

        # Compute kernel matrix
        kernmat = np.exp(-(pdists ** 2) / 2.0)

        return Kernel._reshape_kernelmatrix(kernmat, newshape=kernshape)



import numpy as np
from scipy.integrate._ivp import rk

from probnum import diffeq
from probnum.diffeq import wrappedscipysolver


def setup_solver(y0, ode):
    scipysolver = rk.RK45(ode.rhs, ode.t0, y0, ode.tmax)
    testsolver = wrappedscipysolver.WrappedScipyRungeKutta(
        rk.RK45(ode.rhs, ode.t0, y0, ode.tmax)
    )
    return testsolver, scipysolver


def case_lorenz():
    y0 = np.array([0.0, 1.0, 1.05])
    ode = diffeq.lorenz([0.0, 1.0], y0)
    return setup_solver(y0, ode)


def case_logistic():
    y0 = np.array([0.1])
    ode = diffeq.logistic([0.0, 1.0], y0)
    return setup_solver(y0, ode)


def case_lotkavolterra():
    y0 = np.array([0.1, 0.1])
    ode = diffeq.lotkavolterra([0.0, 1.0], y0)
    return setup_solver(y0, ode)



import unittest

import numpy as np

from probnum.diffeq.ode import ivp
from probnum.randvars import Constant

TEST_NDIM = 3


class TestIVP(unittest.TestCase):
    def setUp(self):
        def rhs_(t, x):
            return -x

        def jac_(t, x):
            return -np.eye(len(x))

        def sol_(t):
            return np.exp(-t) * np.ones(TEST_NDIM)

        some_center = np.random.rand(TEST_NDIM)
        rv = Constant(some_center)
        self.mockivp = ivp.IVP(
            (0.0, np.random.rand()), rv, rhs=rhs_, jac=jac_, sol=sol_
        )

    def test_rhs(self):
        some_x = np.random.rand(TEST_NDIM)
        some_t = np.random.rand()
        out = self.mockivp.rhs(some_t, some_x)
        self.assertEqual(len(out), TEST_NDIM)

    def test_jacobian(self):
        some_x = np.random.rand(TEST_NDIM)
        some_t = np.random.rand()
        out = self.mockivp.jacobian(some_t, some_x)
        self.assertEqual(out.shape[0], TEST_NDIM)
        self.assertEqual(out.shape[1], TEST_NDIM)

    def test_solution(self):
        some_t = np.random.rand()
        out = self.mockivp.solution(some_t)
        self.assertEqual(out.ndim, 1)
        self.assertEqual(out.shape[0], TEST_NDIM)

    def test_initialdistribution(self):
        __ = self.mockivp.initialdistribution

    def test_timespan(self):
        __, __ = self.mockivp.timespan



import numpy as np

from .ivp import IVP


def logistic(timespan, initrv, params=(3.0, 1.0)):
    r"""Initial value problem (IVP) based on the logistic ODE.

    The logistic ODE is defined through

    .. math::

        f(t, y) = a  y  \left( 1 - \frac{y}{b} \right)

    for some parameters :math:`(a, b)`.
    Default is :math:`(a, b)=(3.0, 1.0)`. This implementation includes
    the Jacobian :math:`J_f` of :math:`f` as well as a closed form
    solution given by

    .. math::

        f(t) = \frac{b y_0 \exp(a t)}{b + y_0 \left[ \exp(at) - 1 \right]}

    where :math:`y_0= y(t_0)` is the initial value.

    Parameters
    ----------
    timespan : (float, float)
        Time span of IVP.
    initrv : RandomVariable,
        *(shape=())* -- Scalar-valued RandomVariable that describes the belief
        over the initial value. Usually it is a Constant (noise-free      or Normal (no
        Random Variable isy) with scalar mean and scalar variance.
        To replicate "classical" initial values use the Constant distribution.
    params : (float, float), optional
        Parameters :math:`(a, b)` for the logistic IVP.
        Default is :math:`(a, b) = (3.0, 1.0)`.

    Returns
    -------
    IVP
        IVP object describing the logistic IVP with the prescribed
        configuration.
    """

    def rhs(t, y):
        return log_rhs(t, y, params)

    def jac(t, y):
        return log_jac(t, y, params)

    def hess(t, y):
        return log_hess(t, y, params)

    def sol(t):
        return log_sol(t, params, initrv.mean)

    return IVP(timespan, initrv, rhs, jac, hess, sol)


def log_rhs(t, y, params):
    """RHS for logistic model."""
    l0, l1 = params
    return l0 * y * (1.0 - y / l1)


def log_jac(t, y, params):
    """Jacobian for logistic model."""
    l0, l1 = params
    return np.array([l0 - l0 / l1 * 2 * y])


def log_hess(t, y, params):
    """Hessian for logistic model."""
    l0, l1 = params
    return np.array([[-2 * l0 / l1]])


def log_sol(t, params, y0):
    """Solution for logistic model."""
    l0, l1 = params
    nomin = l1 * y0 * np.exp(l0 * t)
    denom = l1 + y0 * (np.exp(l0 * t) - 1)
    return nomin / denom


def fitzhughnagumo(timespan, initrv, params=(0.0, 0.08, 0.07, 1.25)):
    r"""Initial value problem (IVP) based on the FitzHugh-Nagumo model.

    The FitzHugh-Nagumo (FHN) model is defined through

    .. math::

        f(t, y) =
        \begin{pmatrix}
            y_1 - \frac{1}{3} y_1^3 - y_2 + a \\
            \frac{1}{d} (y_1 + b - c y_2)
        \end{pmatrix}

    for some parameters :math:`(a, b, c, d)`.
    Default is :math:`(a, b)=(0.0, 0.08, 0.07, 1.25)`.
    This implementation includes the Jacobian :math:`J_f` of :math:`f`.

    Parameters
    ----------
    timespan : (float, float)
        Time span of IVP.
    initrv : RandomVariable,
        *(shape=(2, ))* -- Vector-valued RandomVariable that describes the belief
        over the initial value. Usually it is a Constant (noise-free) or Normal (noisy)
        Random Variable with :math:`2`-dimensional mean vector and
        :math:`2 \times 2`-dimensional covariance matrix.
        To replicate "classical" initial values use the Constant distribution.
    params : (float, float, float, float), optional
        Parameters :math:`(a, b, c, d)` for the FitzHugh-Nagumo IVP.
        Default is :math:`(a, b, c, d)=(0.0, 0.08, 0.07, 1.25)`.

    Returns
    -------
    IVP
        IVP object describing the FitzHugh-Nagumo IVP with the prescribed
        configuration.
    """

    def rhs(t, y):
        return fhn_rhs(t, y, params)

    def jac(t, y):
        return fhn_jac(t, y, params)

    return IVP(timespan, initrv, rhs, jac)


def fhn_rhs(t, y, params):
    """RHS for FitzHugh-Nagumo model."""
    y1, y2 = y
    a, b, c, d = params
    return np.array([y1 - y1 ** 3.0 / 3.0 - y2 + a, (y1 + b - c * y2) / d])


def fhn_jac(t, y, params):
    """Jacobian for FitzHugh-Nagumo model."""
    y1, y2 = y
    a, b, c, d = params
    return np.array([[1.0 - y1 ** 2.0, -1.0], [1.0 / d, -c / d]])


def lotkavolterra(timespan, initrv, params=(0.5, 0.05, 0.5, 0.05)):
    r"""Initial value problem (IVP) based on the Lotka-Volterra model.

    The Lotka-Volterra (LV) model is defined through

    .. math::

        f(t, y) =
        \begin{pmatrix}
            a y_1 - by_1y_2 \\
            -c y_2 + d y_1 y_2
        \end{pmatrix}

    for some parameters :math:`(a, b, c, d)`.
    Default is :math:`(a, b)=(0.5, 0.05, 0.5, 0.05)`.
    This implementation includes the Jacobian :math:`J_f` of :math:`f`.

    Parameters
    ----------
    timespan : (float, float)
        Time span of IVP.
    initrv : RandomVariable,
        *(shape=(2, ))* -- Vector-valued RandomVariable that describes the belief
        over the initial value. Usually it is a Constant (noise-free) or Normal (noisy)
        Random Variable with :math:`2`-dimensional mean vector and
        :math:`2 \times 2`-dimensional covariance matrix.
        To replicate "classical" initial values use the Constant distribution.
    params : (float, float, float, float), optional
        Parameters :math:`(a, b, c, d)` for the Lotka-Volterra IVP.
        Default is :math:`(a, b, c, d)=(0.5, 0.05, 0.5, 0.05)`.

    Returns
    -------
    IVP
        IVP object describing the Lotka-Volterra
        IVP with the prescribed
        configuration.
    """

    def rhs(t, y):
        return lv_rhs(t, y, params)

    def jac(t, y):
        return lv_jac(t, y, params)

    return IVP(timespan, initrv, rhs, jac)


def lv_rhs(t, y, params):
    """RHS for Lotka-Volterra."""
    a, b, c, d = params
    y1, y2 = y
    return np.array([a * y1 - b * y1 * y2, -c * y2 + d * y1 * y2])


def lv_jac(t, y, params):
    """Jacobian for Lotka-Volterra."""
    a, b, c, d = params
    y1, y2 = y
    return np.array([[a - b * y2, -b * y1], [d * y2, -c + d * y1]])


def seir(timespan, initrv, params=(0.3, 0.3, 0.1)):
    r"""Initial value problem (IVP) based on the SEIR model.

    The SEIR model with no vital dynamics is defined through

    .. math::

        f(t, y) =
        \begin{pmatrix}
            \frac{-\beta y_1 y_3}{N} \\
            \frac{\beta y_1 y_3}{N} - \alpha y_2 \\
            \alpha y_2 - \gamma y_3 \\
            \gamma y_3
        \end{pmatrix}

    for some parameters :math:`(\alpha, \beta, \gamma)` and population
    count :math:`N`. Without taking vital dynamics into consideration,
    :math:`N` is constant such that for every time point :math:`t`

    .. math::

        S(t) + E(t) + I(t) + R(t) = N

    holds.
    Default parameters are :math:`(\alpha, \beta, \gamma)=(0.3, 0.3, 0.1)`.
    The population count is computed from the (mean of the)
    initial value random variable.
    This implementation includes the Jacobian :math:`J_f` of :math:`f`.

    Parameters
    ----------
    timespan : (float, float)
        Time span of IVP.
    initrv : RandomVariable,
        *(shape=(4, ))* -- Vector-valued RandomVariable that describes the belief
        over the initial value. Usually it is a Constant (noise-free) or Normal (noisy)
        Random Variable with :math:`4`-dimensional mean vector and
        :math:`4 \times 4`-dimensional covariance matrix.
        To replicate "classical" initial values use the Constant distribution.
    params : (float, float, float), optional
        Parameters :math:`(\alpha, \beta, \gamma)` for the SEIR model IVP.
        Default is :math:`(\alpha, \beta, \gamma)=(0.3, 0.3, 0.1)`.

    Returns
    -------
    IVP
        IVP object describing the SEIR model IVP with the prescribed
        configuration.
    """

    population_count = np.sum(initrv.mean)
    params_and_population_count = (*params, population_count)

    def rhs(t, y):
        return seir_rhs(t, y, params_and_population_count)

    def jac(t, y):
        return seir_jac(t, y, params_and_population_count)

    return IVP(timespan, initrv, rhs, jac=jac)


def seir_rhs(t, y, params):
    """RHS for SEIR model."""
    alpha, beta, gamma, population_count = params
    y1, y2, y3, y4 = y
    y1_next = -beta * y1 * y3 / population_count
    y2_next = beta * y1 * y3 / population_count - alpha * y2
    y3_next = alpha * y2 - gamma * y3
    y4_next = gamma * y3

    return np.array([y1_next, y2_next, y3_next, y4_next])


def seir_jac(t, y, params):
    """Jacobian for SEIR model."""
    alpha, beta, gamma, population_count = params
    y1, y2, y3, y4 = y
    d_dy1 = np.array(
        [-beta * y3 / population_count, 0.0, -beta * y1 / population_count, 0.0]
    )
    d_dy2 = np.array(
        [beta * y3 / population_count, -alpha, beta * y1 / population_count, 0.0]
    )
    d_dy3 = np.array([0.0, alpha, -gamma, 0.0])
    d_dy4 = np.array([0.0, 0.0, gamma, 0.0])
    jac_matrix = np.array([d_dy1, d_dy2, d_dy3, d_dy4])
    return jac_matrix


def lorenz(timespan, initrv, params=(10.0, 28.0, 8.0 / 3.0)):
    r"""Initial value problem (IVP) based on the Lorenz system.

    The Lorenz system is defined through

    .. math::

        f(t, y) =
        \begin{pmatrix}
            a(y_2 - y_1) \\
            y_1(b-y_3) - y_2 \\
            y_1y_2 - cy_3
        \end{pmatrix}

    for some parameters :math:`(a, b, c)`.
    Default is :math:`(a, b, c)=(10, 28, 2.667)`.
    This implementation includes the Jacobian :math:`J_f` of :math:`f`.

    Parameters
    ----------
    timespan : (float, float)
        Time span of IVP.
    initrv : RandomVariable,
        *(shape=(3, ))* -- Vector-valued RandomVariable that describes the belief
        over the initial value. Usually it is a Constant (noise-free) or Normal (noisy)
        Random Variable with :math:`3`-dimensional mean vector and
        :math:`3 \times 3`-dimensional covariance matrix.
        To replicate "classical" initial values use the Constant distribution.
    params : (float, float, float, float), optional
        Parameters :math:`(a, b, c)` for the Lorenz system.
        Default is :math:`(a, b, c)=(10, 28, 2.667)`.

    Returns
    -------
    IVP
        IVP object describing the Lorenz system IVP with the prescribed
        configuration.
    """

    def rhs(t, y):
        return lor_rhs(t, y, params)

    def jac(t, y):
        return lor_jac(t, y, params)

    return IVP(timespan, initrv, rhs, jac)


def lor_rhs(t, y, params):
    """RHS for Lorenz system."""
    a, b, c = params
    y1, y2, y3 = y
    return np.array([a * (y2 - y1), y1 * (b - y3) - y2, y1 * y2 - c * y3])


def lor_jac(t, y, params):
    """Jacobian for Lorenz system."""
    a, b, c = params
    y1, y2, y3 = y
    return np.array([[-a, a, 0], [b - y3, -1, -y1], [y2, y1, -c]])



from typing import Optional, Union

import numpy as np

import probnum.utils
from probnum.type import DTypeArgType, ScalarArgType, ShapeArgType

from . import _linear_operator


class Scaling(_linear_operator.LinearOperator):
    def __init__(
        self,
        factors: Union[np.ndarray, ScalarArgType],
        shape: Optional[ShapeArgType] = None,
        dtype: Optional[DTypeArgType] = None,
    ):
        self._factors = None
        self._scalar = None

        if np.ndim(factors) == 0:
            # Isotropic scaling
            self._scalar = probnum.utils.as_numpy_scalar(factors, dtype=dtype)

            if shape is None:
                raise ValueError(
                    "When specifying the scaling factors by a scalar, a shape must be "
                    "specified."
                )

            shape = probnum.utils.as_shape(shape)

            if len(shape) == 1:
                shape = 2 * shape
            elif len(shape) != 2:
                raise ValueError(
                    "The shape of a linear operator must be two-dimensional."
                )

            if shape[0] != shape[1]:
                raise np.linalg.LinAlgError("Scaling operators must be square.")

            dtype = self._scalar.dtype

            if self._scalar == 1:
                # Identity
                matmul = lambda x: x.astype(
                    np.result_type(self.dtype, x.dtype), copy=False
                )
                rmatmul = lambda x: x.astype(
                    np.result_type(self.dtype, x.dtype), copy=False
                )

                apply = lambda x, axis: x.astype(
                    np.result_type(self.dtype, x.dtype), copy=False
                )

                todense = lambda: np.identity(shape[0], dtype=dtype)

                conjugate = lambda: self
                inverse = lambda: self

                rank = lambda: np.intp(shape[0])
                cond = self._cond_isotropic
                eigvals = lambda: np.ones(shape[0], dtype=self._inexact_dtype)
                det = lambda: self._scalar.astype(self._inexact_dtype, copy=False)
                logabsdet = lambda: (0 * self._scalar).astype(
                    self._inexact_dtype, copy=False
                )
            else:
                matmul = lambda x: self._scalar * x
                rmatmul = lambda x: self._scalar * x

                apply = lambda x, axis: self._scalar * x

                todense = self._todense_isotropic

                conjugate = lambda: (
                    self
                    if np.imag(self._scalar) == 0
                    else Scaling(np.conj(self._scalar), shape=shape)
                )
                inverse = self._inverse_isotropic

                rank = lambda: np.intp(0 if self._scalar == 0 else shape[0])
                cond = self._cond_isotropic
                eigvals = lambda: np.full(
                    shape[0], self._scalar, dtype=self._inexact_dtype
                )
                det = lambda: (
                    self._scalar.astype(self._inexact_dtype, copy=False) ** shape[0]
                )
                logabsdet = lambda: (
                    probnum.utils.as_numpy_scalar(-np.inf, dtype=self._inexact_dtype)
                    if self._scalar == 0
                    else shape[0] * np.log(np.abs(self._scalar))
                )

            trace = lambda: self.shape[0] * self._scalar
        elif np.ndim(factors) == 1:
            # Anisotropic scaling
            self._factors = np.asarray(factors, dtype=dtype)
            self._factors.setflags(write=False)

            shape = 2 * self._factors.shape
            dtype = self._factors.dtype

            matmul = lambda x: self._factors[:, np.newaxis] * x
            rmatmul = lambda x: self._factors * x

            apply = lambda x, axis: (
                self._factors.reshape((-1,) + (x.ndim - (axis + 1)) * (1,)) * x
            )

            todense = lambda: np.diag(self._factors)

            conjugate = lambda: (
                self
                if (
                    not np.issubdtype(dtype, np.complexfloating)
                    or np.all(np.imag(self._factors) == 0)
                )
                else Scaling(np.conj(self._factors))
            )
            inverse = self._inverse_anisotropic

            rank = lambda: np.count_nonzero(self.factors, axis=0)
            eigvals = lambda: self._factors
            cond = self._cond_anisotropic
            det = lambda: np.prod(self._factors).astype(self._inexact_dtype, copy=False)
            logabsdet = None
            trace = lambda: np.sum(self._factors)
        else:
            raise TypeError(
                "`factors` must either be a scalar or a 1-dimensional array-like"
            )

        super().__init__(
            shape,
            dtype,
            matmul=matmul,
            rmatmul=rmatmul,
            apply=apply,
            todense=todense,
            conjugate=conjugate,
            transpose=lambda: self,
            adjoint=conjugate,
            inverse=inverse,
            rank=rank,
            eigvals=eigvals,
            cond=cond,
            det=det,
            logabsdet=logabsdet,
            trace=trace,
        )

    @property
    def factors(self) -> np.ndarray:
        if self._factors is None:
            self._factors = np.full(self.shape[0], self._scalar, dtype=self._dtype)

        return self._factors

    @property
    def scalar(self) -> Optional[np.number]:
        return self._scalar

    @property
    def is_isotropic(self) -> bool:
        return self._scalar is not None

    def _astype(self, dtype, order, casting, copy) -> "Scaling":
        if self.dtype == dtype and not copy:
            return self
        else:
            if self.is_isotropic:
                return Scaling(self._scalar, shape=self.shape, dtype=dtype)
            else:
                return Scaling(self._factors, dtype=dtype)

    def _todense_isotropic(self) -> np.ndarray:
        dense = np.zeros(self.shape, dtype=self.dtype)
        np.fill_diagonal(dense, self._scalar)
        return dense

    def _inverse_anisotropic(self) -> "Scaling":
        if self.rank() < self.shape[0]:
            raise np.linalg.LinAlgError("The operator is singular.")

        return Scaling(1 / self._factors)

    def _inverse_isotropic(self) -> "Scaling":
        if self.rank() < self.shape[0]:
            raise np.linalg.LinAlgError("The operator is singular.")

        return Scaling(1 / self._scalar, shape=self.shape)

    def _cond_anisotropic(self, p: Union[None, int, float, str]) -> np.inexact:
        abs_diag = np.abs(self._factors)
        abs_min = np.min(abs_diag)

        if abs_min == 0.0:
            # The operator is singular
            return probnum.utils.as_numpy_scalar(np.inf, dtype=self._inexact_dtype)

        if p is None:
            p = 2

        if p in (2, 1, np.inf, -2, -1, -np.inf):
            abs_max = np.max(abs_diag)

            if p > 0:  # p in (2, 1, np.inf)
                cond = abs_max / abs_min
            else:  # p in (-2, -1, -np.inf)
                if abs_max > 0:
                    cond = abs_min / abs_max
                else:
                    cond = np.double(np.inf)

            return cond.astype(self._inexact_dtype, copy=False)
        elif p == "fro":
            norm = np.linalg.norm(self._factors, ord=2)
            norm_inv = np.linalg.norm(1 / self._factors, ord=2)
            return (norm * norm_inv).astype(self._inexact_dtype, copy=False)

        return np.linalg.cond(self.todense(cache=False), p=p)

    def _cond_isotropic(self, p: Union[None, int, float, str]) -> np.inexact:
        if self._scalar == 0:
            return self._inexact_dtype.type(np.inf)

        if p is None or p in (2, 1, np.inf, -2, -1, -np.inf):
            return probnum.utils.as_numpy_scalar(1.0, dtype=self._inexact_dtype)
        elif p == "fro":
            return probnum.utils.as_numpy_scalar(
                min(self.shape), dtype=self._inexact_dtype
            )
        else:
            return np.linalg.cond(self.todense(cache=False), p=p)



from typing import Tuple

import numpy as np
import pytest

import probnum as pn

matrices = [
    np.array([[-1.5, 3], [0, -230]]),
    np.array([[2, 0], [1, 3]]),
    np.array([[2, 0, -1.5], [1, 3, -230]]),
]


@pytest.mark.parametrize("matrix", matrices)
def case_matvec(matrix: np.ndarray) -> Tuple[pn.linops.LinearOperator, np.ndarray]:
    @pn.linops.LinearOperator.broadcast_matvec
    def _matmul(vec: np.ndarray):
        return matrix @ vec

    linop = pn.linops.LinearOperator(
        shape=matrix.shape, dtype=matrix.dtype, matmul=_matmul
    )

    return linop, matrix


@pytest.mark.parametrize("matrix", matrices)
def case_matrix(matrix: np.ndarray) -> Tuple[pn.linops.LinearOperator, np.ndarray]:
    return pn.linops.Matrix(matrix), matrix


@pytest.mark.parametrize("n", [3, 4, 8, 12, 15])
def case_identity(n: int) -> Tuple[pn.linops.LinearOperator, np.ndarray]:
    return pn.linops.Identity(shape=n), np.eye(n)



import numpy as np
import pytest

import probnum.statespace as pnss
from probnum import randvars
from probnum.problems.zoo.linalg import random_spd_matrix

from .test_sde import TestLTISDE


@pytest.fixture
def some_ordint(test_ndim):
    return test_ndim - 1


class TestIntegrator:
    """An integrator should be usable as is, but its tests are also useful for IBM,
    IOUP, etc."""

    # Replacement for an __init__ in the pytest language. See:
    # https://stackoverflow.com/questions/21430900/py-test-skips-test-class-if-constructor-is-defined
    @pytest.fixture(autouse=True)
    def _setup(self, some_ordint):
        self.some_ordint = some_ordint
        self.integrator = pnss.Integrator(ordint=self.some_ordint, spatialdim=1)

    def test_proj2coord(self):
        base = np.zeros(self.some_ordint + 1)
        base[0] = 1
        e_0_expected = np.kron(np.eye(1), base)
        e_0 = self.integrator.proj2coord(coord=0)
        np.testing.assert_allclose(e_0, e_0_expected)

        base = np.zeros(self.some_ordint + 1)
        base[-1] = 1
        e_q_expected = np.kron(np.eye(1), base)
        e_q = self.integrator.proj2coord(coord=self.some_ordint)
        np.testing.assert_allclose(e_q, e_q_expected)

    def test_precon(self):

        assert isinstance(self.integrator.precon, pnss.NordsieckLikeCoordinates)


class TestIBM(TestLTISDE, TestIntegrator):

    # Replacement for an __init__ in the pytest language. See:
    # https://stackoverflow.com/questions/21430900/py-test-skips-test-class-if-constructor-is-defined
    @pytest.fixture(autouse=True)
    def _setup(
        self,
        some_ordint,
        forw_impl_string_linear_gauss,
        backw_impl_string_linear_gauss,
    ):
        self.some_ordint = some_ordint
        spatialdim = 1  # make tests compatible with some_normal_rv1, etc.
        self.transition = pnss.IBM(
            ordint=self.some_ordint,
            spatialdim=spatialdim,
            forward_implementation=forw_impl_string_linear_gauss,
            backward_implementation=backw_impl_string_linear_gauss,
        )

        self.G = lambda t: self.transition.driftmat
        self.v = lambda t: self.transition.forcevec
        self.L = lambda t: self.transition.dispmat

        self.g = lambda t, x: self.G(t) @ x + self.v(t)
        self.dg = lambda t, x: self.G(t)

    @property
    def integrator(self):
        return self.transition


class TestIOUP(TestLTISDE, TestIntegrator):

    # Replacement for an __init__ in the pytest language. See:
    # https://stackoverflow.com/questions/21430900/py-test-skips-test-class-if-constructor-is-defined
    @pytest.fixture(autouse=True)
    def _setup(
        self,
        some_ordint,
        forw_impl_string_linear_gauss,
        backw_impl_string_linear_gauss,
    ):
        self.some_ordint = some_ordint
        spatialdim = 1  # make tests compatible with some_normal_rv1, etc.
        self.transition = pnss.IOUP(
            ordint=self.some_ordint,
            spatialdim=spatialdim,
            driftspeed=1.2345,
            forward_implementation=forw_impl_string_linear_gauss,
            backward_implementation=backw_impl_string_linear_gauss,
        )

        self.G = lambda t: self.transition.driftmat
        self.v = lambda t: self.transition.forcevec
        self.L = lambda t: self.transition.dispmat

        self.g = lambda t, x: self.G(t) @ x + self.v(t)
        self.dg = lambda t, x: self.G(t)

    @property
    def integrator(self):
        return self.transition


class TestMatern(TestLTISDE, TestIntegrator):

    # Replacement for an __init__ in the pytest language. See:
    # https://stackoverflow.com/questions/21430900/py-test-skips-test-class-if-constructor-is-defined
    @pytest.fixture(autouse=True)
    def _setup(
        self,
        some_ordint,
        forw_impl_string_linear_gauss,
        backw_impl_string_linear_gauss,
    ):
        self.some_ordint = some_ordint
        spatialdim = 1  # make tests compatible with some_normal_rv1, etc.
        self.transition = pnss.Matern(
            ordint=self.some_ordint,
            spatialdim=spatialdim,
            lengthscale=1.2345,
            forward_implementation=forw_impl_string_linear_gauss,
            backward_implementation=backw_impl_string_linear_gauss,
        )

        self.G = lambda t: self.transition.driftmat
        self.v = lambda t: self.transition.forcevec
        self.L = lambda t: self.transition.dispmat

        self.g = lambda t, x: self.G(t) @ x + self.v(t)
        self.dg = lambda t, x: self.G(t)

    @property
    def integrator(self):
        return self.transition


def both_transitions_matern():
    matern = pnss.Matern(ordint=2, spatialdim=2, lengthscale=2.041)
    matern2 = pnss.Matern(ordint=2, spatialdim=2, lengthscale=2.041)
    matern_as_ltisde = pnss.LTISDE(matern2.driftmat, matern2.forcevec, matern2.dispmat)
    return matern, matern_as_ltisde


def both_transitions_ioup():
    ioup = pnss.IOUP(ordint=2, spatialdim=2, driftspeed=2.041)
    ioup2 = pnss.IOUP(ordint=2, spatialdim=2, driftspeed=2.041)
    ioup_as_ltisde = pnss.LTISDE(ioup2.driftmat, ioup2.forcevec, ioup2.dispmat)
    return ioup, ioup_as_ltisde


def both_transitions_ibm():
    ibm = pnss.IBM(ordint=2, spatialdim=1)
    ibm2 = pnss.IBM(ordint=2, spatialdim=1)
    ibm_as_ltisde = pnss.LTISDE(ibm2.driftmat, ibm2.forcevec, ibm2.dispmat)
    return ibm, ibm_as_ltisde


@pytest.mark.parametrize(
    "both_transitions",
    [both_transitions_ibm(), both_transitions_ioup(), both_transitions_matern()],
)
def test_same_forward_outputs(both_transitions, diffusion):
    trans1, trans2 = both_transitions
    real = 1 + 0.1 * np.random.rand(trans1.dimension)
    out_1, info1 = trans1.forward_realization(
        real, t=0.0, dt=0.5, compute_gain=True, _diffusion=diffusion
    )
    out_2, info2 = trans2.forward_realization(
        real, t=0.0, dt=0.5, compute_gain=True, _diffusion=diffusion
    )
    np.testing.assert_allclose(out_1.mean, out_2.mean)
    np.testing.assert_allclose(out_1.cov, out_2.cov)
    np.testing.assert_allclose(info1["crosscov"], info2["crosscov"])
    np.testing.assert_allclose(info1["gain"], info2["gain"])


@pytest.mark.parametrize(
    "both_transitions",
    [both_transitions_ibm(), both_transitions_ioup(), both_transitions_matern()],
)
def test_same_backward_outputs(both_transitions, diffusion):
    trans1, trans2 = both_transitions
    real = 1 + 0.1 * np.random.rand(trans1.dimension)
    real2 = 1 + 0.1 * np.random.rand(trans1.dimension)
    cov = random_spd_matrix(trans1.dimension)
    rv = randvars.Normal(real2, cov)
    out_1, info1 = trans1.backward_realization(
        real, rv, t=0.0, dt=0.5, compute_gain=True, _diffusion=diffusion
    )
    out_2, info2 = trans2.backward_realization(
        real, rv, t=0.0, dt=0.5, compute_gain=True, _diffusion=diffusion
    )
    np.testing.assert_allclose(out_1.mean, out_2.mean)
    np.testing.assert_allclose(out_1.cov, out_2.cov)

    # Both dicts are empty?
    assert not info1
    assert not info2


@pytest.fixture
def dt():
    return 0.1


@pytest.fixture
def ah_22_ibm(dt):
    return np.array(
        [
            [1.0, dt, dt ** 2 / 2.0],
            [0.0, 1.0, dt],
            [0.0, 0.0, 1.0],
        ]
    )


@pytest.fixture
def qh_22_ibm(dt):
    return np.array(
        [
            [dt ** 5 / 20.0, dt ** 4 / 8.0, dt ** 3 / 6.0],
            [dt ** 4 / 8.0, dt ** 3 / 3.0, dt ** 2 / 2.0],
            [dt ** 3 / 6.0, dt ** 2 / 2.0, dt],
        ]
    )


@pytest.fixture
def spdmat3x3():
    return random_spd_matrix(3)


@pytest.fixture
def normal_rv3x3(spdmat3x3):

    return randvars.Normal(
        mean=np.random.rand(3),
        cov=spdmat3x3,
        cov_cholesky=np.linalg.cholesky(spdmat3x3),
    )


class TestIBMValues:

    # Replacement for an __init__ in the pytest language. See:
    # https://stackoverflow.com/questions/21430900/py-test-skips-test-class-if-constructor-is-defined
    @pytest.fixture(autouse=True)
    def _setup(
        self,
        forw_impl_string_linear_gauss,
        backw_impl_string_linear_gauss,
    ):
        spatialdim = 1  # make tests compatible with some_normal_rv1, etc.
        self.transition = pnss.IBM(
            ordint=2,
            spatialdim=spatialdim,
            forward_implementation=forw_impl_string_linear_gauss,
            backward_implementation=backw_impl_string_linear_gauss,
        )

    def test_discretise_values(self, ah_22_ibm, qh_22_ibm, dt):
        discrete_model = self.transition.discretise(dt=dt)
        np.testing.assert_allclose(discrete_model.state_trans_mat, ah_22_ibm)
        np.testing.assert_allclose(discrete_model.proc_noise_cov_mat, qh_22_ibm)

    def test_forward_rv_values(self, normal_rv3x3, diffusion, ah_22_ibm, qh_22_ibm, dt):
        rv, _ = self.transition.forward_rv(
            normal_rv3x3, t=0.0, dt=dt, _diffusion=diffusion
        )
        np.testing.assert_allclose(ah_22_ibm @ normal_rv3x3.mean, rv[:3].mean)
        np.testing.assert_allclose(
            ah_22_ibm @ normal_rv3x3.cov @ ah_22_ibm.T + diffusion * qh_22_ibm,
            rv.cov,
        )

    def test_forward_realization_values(
        self, normal_rv3x3, diffusion, ah_22_ibm, qh_22_ibm, dt
    ):
        real = normal_rv3x3.sample()
        rv, _ = self.transition.forward_realization(
            real, t=0.0, dt=dt, _diffusion=diffusion
        )
        np.testing.assert_allclose(ah_22_ibm @ real, rv.mean)
        np.testing.assert_allclose(diffusion * qh_22_ibm, rv.cov)


############################################################################################
# Tests for the coordinate-representation conversion functions.
############################################################################################


@pytest.fixture
def some_order():
    return 2


@pytest.fixture
def some_dim():
    return 3


@pytest.fixture
def fake_state(some_order, some_dim):
    return np.arange(some_dim * (some_order + 1))


@pytest.fixture
def in_out_pair(fake_state, some_order):
    """Initial states for the three-body initial values.

    Returns (derivwise, coordwise)
    """
    return fake_state, fake_state.reshape((-1, some_order + 1)).T.flatten()


def test_in_out_pair_is_not_identical(in_out_pair):
    """A little sanity check to assert that these are actually different, so the
    conversion is non-trivial."""
    derivwise, coordwise = in_out_pair
    assert np.linalg.norm(derivwise - coordwise) > 5


def test_convert_coordwise_to_derivwise(in_out_pair, some_order, some_dim):
    derivwise, coordwise = in_out_pair
    coordwise_as_derivwise = pnss.Integrator._convert_coordwise_to_derivwise(
        coordwise, some_order, some_dim
    )
    np.testing.assert_allclose(coordwise_as_derivwise, derivwise)


def test_convert_derivwise_to_coordwise(in_out_pair, some_order, some_dim):
    derivwise, coordwise = in_out_pair
    derivwise_as_coordwise = pnss.Integrator._convert_derivwise_to_coordwise(
        derivwise, some_order, some_dim
    )
    np.testing.assert_allclose(derivwise_as_coordwise, coordwise)


def test_conversion_pairwise_inverse(in_out_pair, some_order, some_dim):
    derivwise, coordwise = in_out_pair
    as_coord = pnss.Integrator._convert_derivwise_to_coordwise(
        derivwise, some_order, some_dim
    )
    as_deriv_again = pnss.Integrator._convert_coordwise_to_derivwise(
        as_coord, some_order, some_dim
    )
    np.testing.assert_allclose(as_deriv_again, derivwise)



from typing import Tuple

import numpy as np
import pytest

import probnum as pn


@pytest.mark.parametrize(
    "diagonal",
    [
        np.array([1.0, 2, -3]),
        np.array([1.0, 2.0, 0.0, -1.0]),
    ],
)
def case_anisotropic_scaling(
    diagonal: np.ndarray,
) -> Tuple[pn.linops.LinearOperator, np.ndarray]:
    return pn.linops.Scaling(diagonal), np.diag(diagonal)


@pytest.mark.parametrize("n", [3, 4, 8, 12, 15])
@pytest.mark.parametrize(
    "scalar",
    [
        0.0,  # Singular matrix
        1.0,  # Identity
        5,  # Integer dtype
        -8.3,
    ],
)
def case_isotropic_scaling(
    n: int, scalar
) -> Tuple[pn.linops.LinearOperator, np.ndarray]:
    return pn.linops.Scaling(scalar, shape=n), np.diag(np.full((n,), scalar))



"""Test cases for Gaussian Filtering and Smoothing."""
import unittest

import numpy as np

import probnum.problems.zoo.filtsmooth as filtsmooth_zoo
from probnum import filtsmooth
from tests.testing import NumpyAssertions

__all__ = [
    "LinearisedDiscreteTransitionTestCase",
]

# Show plots in tests?
VISUALISE = False

if VISUALISE:
    try:
        import matplotlib.pyplot as plt
    except ImportError as err:
        raise ImportError(
            "Install matplotlib to visualise the test functions."
        ) from err


np.random.seed(12345)


class LinearisedDiscreteTransitionTestCase(unittest.TestCase, NumpyAssertions):
    """Test approximate Gaussian filtering and smoothing.

    1. Transition RV is enabled by linearising
    2. Applied to a linear model, the outcome is exact
    3. Smoothing RMSE < Filtering RMSE < Data RMSE on the pendulum example.
    """

    linearising_component_pendulum = NotImplemented
    linearising_component_car = NotImplemented

    def test_transition_rv(self):
        """forward_rv() not possible for original model but for the linearised model."""
        # pylint: disable=not-callable
        _, statespace_components = filtsmooth_zoo.pendulum()
        nonlinear_model = statespace_components["dynamics_model"]
        initrv = statespace_components["initrv"]
        linearised_model = self.linearising_component_pendulum(nonlinear_model)

        with self.subTest("Baseline should not work."):
            with self.assertRaises(NotImplementedError):
                nonlinear_model.forward_rv(initrv, 0.0)
        with self.subTest("Linearisation happens."):
            linearised_model.forward_rv(initrv, 0.0)

    def test_exactness_linear_model(self):
        """Applied to a linear model, the results should be unchanged."""
        # pylint: disable=not-callable
        regression_problem, statespace_components = filtsmooth_zoo.car_tracking()
        linear_model = statespace_components["dynamics_model"]
        initrv = statespace_components["initrv"]
        linearised_model = self.linearising_component_car(linear_model)

        with self.subTest("Different objects"):
            self.assertNotIsInstance(linear_model, type(linearised_model))

        received, info1 = linear_model.forward_rv(initrv, 0.0)
        expected, info2 = linearised_model.forward_rv(initrv, 0.0)
        crosscov1 = info1["crosscov"]
        crosscov2 = info2["crosscov"]
        rtol, atol = 1e-10, 1e-10
        with self.subTest("Same outputs"):
            self.assertAllClose(received.mean, expected.mean, rtol=rtol, atol=atol)
            self.assertAllClose(received.cov, expected.cov, rtol=rtol, atol=atol)
            self.assertAllClose(crosscov1, crosscov2, rtol=rtol, atol=atol)

    def test_filtsmooth_pendulum(self):
        # pylint: disable=not-callable
        # Set up test problem
        regression_problem, statespace_components = filtsmooth_zoo.pendulum()

        # Linearise problem
        ekf_meas = self.linearising_component_pendulum(
            statespace_components["measurement_model"]
        )
        ekf_dyna = self.linearising_component_pendulum(
            statespace_components["dynamics_model"]
        )
        method = filtsmooth.Kalman(ekf_dyna, ekf_meas, statespace_components["initrv"])

        # Compute filter/smoother solution
        posterior, _ = method.filtsmooth(regression_problem)
        filtms = posterior.filtering_posterior.states.mean
        smooms = posterior.states.mean

        # Compute RMSEs
        comp = regression_problem.solution[:, 0]
        normaliser = np.sqrt(comp.size)
        filtrmse = np.linalg.norm(filtms[:, 0] - comp) / normaliser
        smoormse = np.linalg.norm(smooms[:, 0] - comp) / normaliser
        obs_rmse = (
            np.linalg.norm(regression_problem.observations[:, 0] - comp) / normaliser
        )

        # If desired, visualise.
        if VISUALISE:
            obs, tms, states = (
                regression_problem.observations,
                regression_problem.locations,
                regression_problem.solution,
            )
            fig, (ax1, ax2) = plt.subplots(1, 2)
            fig.suptitle(
                "Noisy pendulum model (%.2f " % smoormse
                + "< %.2f < %.2f?)" % (filtrmse, obs_rmse)
            )
            ax1.set_title("Horizontal position")
            ax1.plot(tms[1:], obs[:, 0], ".", alpha=0.25, label="Observations")
            ax1.plot(
                tms[1:],
                np.sin(states)[1:, 0],
                "-",
                linewidth=4,
                alpha=0.5,
                label="Truth",
            )
            ax1.plot(tms[1:], np.sin(filtms)[1:, 0], "-", label="Filter")
            ax1.plot(tms[1:], np.sin(smooms)[1:, 0], "-", label="Smoother")
            ax1.set_xlabel("time")
            ax1.set_ylabel("horizontal pos. = sin(angular)")
            ax1.legend()

            ax2.set_title("Angular position")
            ax2.plot(
                tms[1:],
                states[1:, 0],
                "-",
                linewidth=4,
                alpha=0.5,
                label="Truth",
            )
            ax2.plot(tms[1:], filtms[1:, 0], "-", label="Filter")
            ax2.plot(tms[1:], smooms[1:, 0], "-", label="Smoother")
            ax2.set_xlabel("time")
            ax2.set_ylabel("angular pos.")
            ax2.legend()
            plt.show()

        # Test if RMSEs behave well.
        self.assertLess(smoormse, filtrmse)
        self.assertLess(filtrmse, obs_rmse)


class LinearisedContinuousTransitionTestCase(unittest.TestCase, NumpyAssertions):
    """Test approximate Gaussian filtering and smoothing.

    1. Transition RV is enabled by linearising
    2. Applied to a linear model, the outcome is exact
    3. Smoothing RMSE < Filtering RMSE < Data RMSE on the Benes-Daum example.
    """

    linearising_component_benes_daum = NotImplemented

    def test_transition_rv(self):
        """forward_rv() not possible for original model but for the linearised model."""
        # pylint: disable=not-callable
        _, statespace_components = filtsmooth_zoo.benes_daum()
        nonlinear_model = statespace_components["dynamics_model"]
        initrv = statespace_components["initrv"]
        linearised_model = self.linearising_component_benes_daum(nonlinear_model)

        with self.subTest("Baseline should not work."):
            with self.assertRaises(NotImplementedError):
                nonlinear_model.forward_rv(initrv, 0.0, 1.0)
        with self.subTest("Linearisation happens."):
            linearised_model.forward_rv(initrv, 0.0, 1.0)

    def test_transition_real(self):
        """transition_real() not possible for original model but for the linearised
        model."""
        # pylint: disable=not-callable
        _, statespace_components = filtsmooth_zoo.benes_daum()
        nonlinear_model = statespace_components["dynamics_model"]
        initrv = statespace_components["initrv"]
        linearised_model = self.linearising_component_benes_daum(nonlinear_model)

        with self.subTest("Baseline should not work."):
            with self.assertRaises(NotImplementedError):
                nonlinear_model.forward_realization(initrv.mean, 0.0, 1.0)
        with self.subTest("Linearisation happens."):
            linearised_model.forward_realization(initrv.mean, 0.0, 1.0)



"""Exemplary state space model setups for Bayesian Filtering and Smoothing."""


from ._filtsmooth_problems import (
    benes_daum,
    car_tracking,
    logistic_ode,
    ornstein_uhlenbeck,
    pendulum,
)

__all__ = [
    "benes_daum",
    "car_tracking",
    "logistic_ode",
    "ornstein_uhlenbeck",
    "pendulum",
]



"""Previously, this module contained the tests for functions in the
`diffeq.odefiltsmooth.ivp2filter` module, since this module has become obsolete, we test
its replacement (`GaussianIVPFilter.string_to_measurement_model`) here.

They need different fixtures anyway.
"""

import numpy as np
import pytest

from probnum import diffeq, filtsmooth, randvars, statespace


@pytest.fixture
def ivp():
    y0 = randvars.Constant(np.array([20.0, 15.0]))
    return diffeq.lotkavolterra([0.4124, 1.15124], y0)


@pytest.fixture
def prior(ivp):
    ode_dim = ivp.dimension
    prior = statespace.IBM(ordint=2, spatialdim=ode_dim)
    return prior


@pytest.mark.parametrize(
    "string, expected_type",
    [
        ("EK0", filtsmooth.DiscreteEKFComponent),
        ("EK1", filtsmooth.DiscreteEKFComponent),
        ("UK", filtsmooth.DiscreteUKFComponent),
    ],
)
def test_output_type(string, expected_type, ivp, prior):
    """Assert that the output type matches."""
    received = diffeq.GaussianIVPFilter.string_to_measurement_model(string, ivp, prior)
    assert isinstance(received, expected_type)


@pytest.mark.parametrize(
    "string",
    ["EK0", "EK1", "UK"],
)
def test_true_mean_ek(string, ivp, prior):
    """Assert that a forwarded realization is x[1] - f(t, x[0]) with zero added covariance."""
    received = diffeq.GaussianIVPFilter.string_to_measurement_model(string, ivp, prior)
    some_real = 1.0 + 0.01 * np.random.rand(prior.dimension)
    some_time = 1.0 + 0.01 * np.random.rand()
    received, _ = received.forward_realization(some_real, some_time)

    e0, e1 = prior.proj2coord(0), prior.proj2coord(1)
    expected = e1 @ some_real - ivp.rhs(some_time, e0 @ some_real)
    np.testing.assert_allclose(received.mean, expected)
    np.testing.assert_allclose(received.cov, 0.0, atol=1e-12)



"""Wrapper class of scipy.integrate. for RK23 and RK45.

Dense-output can not be used for DOP853, if you use other RK-methods,
make sure, that the current implementation works for them   .
"""
import numpy as np
from scipy.integrate._ivp import rk
from scipy.integrate._ivp.common import OdeSolution

from probnum import diffeq, randvars
from probnum.diffeq import wrappedscipyodesolution
from probnum.type import FloatArgType


class WrappedScipyRungeKutta(diffeq.ODESolver):
    """Wrappper for Runge-Kutta methods from Scipy, implements the stepfunction and
    dense output."""

    def __init__(self, solver: rk.RungeKutta):
        self.solver = solver
        self.interpolants = None

        # ProbNum ODESolver needs an ivp
        ivp = diffeq.IVP(
            timespan=[self.solver.t, self.solver.t_bound],
            initrv=randvars.Constant(self.solver.y),
            rhs=self.solver._fun,
        )

        # Dopri853 as implemented in SciPy computes the dense output differently.
        if isinstance(solver, rk.DOP853):
            raise TypeError(
                "Dense output interpolation of DOP853 is currently not supported. Choose a different RK-method."
            )
        super().__init__(ivp=ivp, order=solver.order)

    def initialise(self):
        """Return t0 and y0 (for the solver, which might be different to ivp.y0) and
        initialize the solver. Reset the solver when solving the ODE multiple times,
        i.e. explicitly setting y_old, t, y and f to the respective initial values,
        otherwise those are wrong when running the solver twice.

        Returns
        -------
        self.ivp.t0: float
            initial time point
        self.ivp.initrv: randvars.RandomVariable
            initial random variable
        """

        self.interpolants = []
        self.solver.y_old = None
        self.solver.t = self.ivp.t0
        self.solver.y = self.ivp.initrv.mean
        self.solver.f = self.solver.fun(self.solver.t, self.solver.y)
        return self.ivp.t0, self.ivp.initrv

    def step(
        self, start: FloatArgType, stop: FloatArgType, current: randvars, **kwargs
    ):
        """Perform one ODE-step from start to stop and set variables to the
        corresponding values.

        To specify start and stop directly, rk_step() and not _step_impl() is used.

        Parameters
        ----------
        start : float
            starting location of the step
        stop : float
            stopping location of the step
        current : :obj:`list` of :obj:`RandomVariable`
            current state of the ODE.

        Returns
        -------
        random_var : randvars.RandomVariable
            Estimated states of the discrete-time solution.
        error_estimation : float
            estimated error after having performed the step.
        """

        y = current.mean
        dt = stop - start
        y_new, f_new = rk.rk_step(
            self.solver.fun,
            start,
            y,
            self.solver.f,
            dt,
            self.solver.A,
            self.solver.B,
            self.solver.C,
            self.solver.K,
        )

        # Unnormalized error estimation is used as the error estimation is normalized in
        # solve().
        error_estimation = self.solver._estimate_error(self.solver.K, dt)
        y_new_as_rv = randvars.Constant(y_new)

        # Update the solver settings. This part is copied from scipy's _step_impl().
        self.solver.h_previous = dt
        self.solver.y_old = current
        self.solver.t_old = start
        self.solver.t = stop
        self.solver.y = y_new
        self.solver.h_abs = dt
        self.solver.f = f_new
        return y_new_as_rv, error_estimation

    def rvlist_to_odesol(self, times: np.array, rvs: np.array):
        """Create a ScipyODESolution object which is a subclass of
        diffeq.ODESolution."""
        scipy_solution = OdeSolution(times, self.interpolants)
        probnum_solution = wrappedscipyodesolution.WrappedScipyODESolution(
            scipy_solution, rvs
        )
        return probnum_solution

    def method_callback(self, time, current_guess, current_error):
        """Call dense output after each step and store the interpolants."""
        dense = self.dense_output()
        self.interpolants.append(dense)

    def dense_output(self):
        """Compute the interpolant after each step.

        Returns
        -------
        sol : rk.RkDenseOutput
            Interpolant between the last and current location.
        """
        Q = self.solver.K.T.dot(self.solver.P)
        sol = rk.RkDenseOutput(
            self.solver.t_old, self.solver.t, self.solver.y_old.mean, Q
        )
        return sol



"""Tests for random processes."""

import numpy as np
import pytest

from probnum import randprocs, randvars

# pylint: disable=invalid-name


def test_output_shape(random_process: randprocs.RandomProcess, args0: np.ndarray):
    """Test whether evaluations of the random process have the correct shape."""
    if random_process.output_dim == 1:
        assert random_process(args0).ndim == 1
    else:
        assert random_process(args0).shape[1] == random_process.output_dim


def test_mean_shape(random_process: randprocs.RandomProcess, args0: np.ndarray):
    """Test whether the mean of the random process has the correct shape."""
    if random_process.output_dim == 1:
        assert random_process.mean(args0).ndim == 1
    else:
        assert random_process.mean(args0).shape[1] == random_process.output_dim


def test_var_shape(random_process: randprocs.RandomProcess, args0: np.ndarray):
    """Test whether the variance of the random process has the correct shape."""
    if random_process.output_dim == 1:
        assert random_process.var(args0).ndim == 1
    else:
        assert random_process.var(args0).shape[1] == random_process.output_dim


def test_std_shape(random_process: randprocs.RandomProcess, args0: np.ndarray):
    """Test whether the standard deviation of the random process has the correct
    shape."""
    if random_process.output_dim == 1:
        assert random_process.std(args0).ndim == 1
    else:
        assert random_process.std(args0).shape[1] == random_process.output_dim


def test_cov_shape(random_process: randprocs.RandomProcess, args0: np.ndarray):
    """Test whether the covariance of the random process has the correct shape."""
    n = args0.shape[0]
    if random_process.output_dim == 1:
        assert (
            random_process.cov(args0).shape == (n, n)
            or random_process.cov(args0).ndim < 2
        )
    else:
        assert random_process.cov(args0).shape == (
            n,
            n,
            random_process.output_dim,
            random_process.output_dim,
        )


def test_evaluated_random_process_is_random_variable(
    random_process: randprocs.RandomProcess, random_state: np.random.RandomState
):
    """Test whether evaluating a random process returns a random variable."""
    n_inputs_args0 = 10
    args0 = random_state.normal(size=(n_inputs_args0, random_process.input_dim))
    y0 = random_process(args0)

    assert isinstance(y0, randvars.RandomVariable), (
        f"Output of {repr(random_process)} is not a " f"random variable."
    )


@pytest.mark.xfail(reason="Not yet implemented for random processes.")
def test_samples_are_callables(
    random_process: randprocs.RandomProcess, random_state: np.random.RandomState
):
    """When not specifying inputs to the sample method it should return ``size`` number
    of callables."""
    assert callable(random_process.sample(random_state=random_state))


@pytest.mark.xfail(reason="Not yet implemented for random processes.")
def test_sample_paths_are_deterministic_functions(
    random_process: randprocs.RandomProcess, args0: np.ndarray
):
    """When sampling paths from a random process, repeated evaluation of the sample path
    at the same inputs should return the same values."""
    sample_path = random_process.sample()
    np.testing.assert_array_equal(sample_path(args0), sample_path(args0))


def test_rp_mean_cov_evaluated_matches_rv_mean_cov(
    random_process: randprocs.RandomProcess, random_state: np.random.RandomState
):
    """Check whether the evaluated mean and covariance function of a random process is
    equivalent to the mean and covariance of the evaluated random process as a random
    variable."""
    x = random_state.normal(size=(10, random_process.input_dim))

    np.testing.assert_allclose(
        random_process(x).mean,
        random_process.mean(x),
        err_msg=f"Mean of evaluated {repr(random_process)} does not match the "
        f"random process mean function evaluated.",
    )

    np.testing.assert_allclose(
        random_process(x).cov,
        random_process.cov(x),
        err_msg=f"Covariance of evaluated {repr(random_process)} does not match the "
        f"random process mean function evaluated.",
    )



"""Gaussian processes."""

from typing import Callable, Optional, Type, Union

import numpy as np

from probnum import kernels, randvars
from probnum.type import RandomStateArgType, ShapeArgType

from . import _random_process

_InputType = Union[np.floating, np.ndarray]
_OutputType = Union[np.floating, np.ndarray]


class GaussianProcess(_random_process.RandomProcess[_InputType, _OutputType]):
    """Gaussian processes.

    A Gaussian process is a continuous stochastic process which if evaluated at a
    finite set of inputs returns a random variable with a normal distribution. Gaussian
    processes are fully characterized by their mean and covariance function.

    Parameters
    ----------
    mean :
        Mean function.
    cov :
        Covariance function or kernel.

    See Also
    --------
    RandomProcess : Random processes.
    MarkovProcess : Random processes with the Markov property.

    Examples
    --------
    Define a Gaussian process with a zero mean function and RBF kernel.

    >>> import numpy as np
    >>> from probnum.kernels import ExpQuad
    >>> from probnum.randprocs import GaussianProcess
    >>> mu = lambda x : np.zeros_like(x)  # zero-mean function
    >>> k = ExpQuad(input_dim=1)  # RBF kernel
    >>> gp = GaussianProcess(mu, k)

    Sample from the Gaussian process.

    >>> x = np.linspace(-1, 1, 5)[:, None]
    >>> np.random.seed(42)
    >>> gp.sample(x)
    array([[-0.35187364],
           [-0.41301096],
           [-0.65094306],
           [-0.56817194],
           [ 0.01173088]])
    >>> gp.cov(x)
    array([[1.        , 0.8824969 , 0.60653066, 0.32465247, 0.13533528],
           [0.8824969 , 1.        , 0.8824969 , 0.60653066, 0.32465247],
           [0.60653066, 0.8824969 , 1.        , 0.8824969 , 0.60653066],
           [0.32465247, 0.60653066, 0.8824969 , 1.        , 0.8824969 ],
           [0.13533528, 0.32465247, 0.60653066, 0.8824969 , 1.        ]])
    """

    def __init__(
        self,
        mean: Callable[[_InputType], _OutputType],
        cov: kernels.Kernel,
    ):
        if not isinstance(cov, kernels.Kernel):
            raise TypeError(
                "The covariance functions must be implemented as a " "`Kernel`."
            )

        self._meanfun = mean
        self._covfun = cov
        super().__init__(
            input_dim=cov.input_dim,
            output_dim=cov.output_dim,
            dtype=np.dtype(np.float_),
        )

    def __call__(self, args: _InputType) -> randvars.Normal:
        return randvars.Normal(mean=self.mean(args), cov=self.cov(args))

    def mean(self, args: _InputType) -> _OutputType:
        return self._meanfun(args)

    def cov(self, args0: _InputType, args1: Optional[_InputType] = None) -> _OutputType:
        return self._covfun(args0, args1)

    def _sample_at_input(
        self,
        args: _InputType,
        size: ShapeArgType = (),
        random_state: RandomStateArgType = None,
    ) -> _OutputType:
        gaussian_rv = self.__call__(args)
        gaussian_rv.random_state = random_state
        return gaussian_rv.sample(size=size)

    def push_forward(
        self,
        args: _InputType,
        base_measure: Type[randvars.RandomVariable],
        sample: np.ndarray,
    ) -> np.ndarray:
        raise NotImplementedError



"""Tests for the normal distribution."""
import itertools
import unittest

import numpy as np
import scipy.linalg
import scipy.sparse
import scipy.stats

import probnum
from probnum import linops, randvars
from probnum.problems.zoo.linalg import random_spd_matrix
from tests.testing import NumpyAssertions


class NormalTestCase(unittest.TestCase, NumpyAssertions):
    """General test case for the normal distribution."""

    def setUp(self):
        """Resources for tests."""
        # Seed
        np.random.seed(seed=42)

        # Parameters
        m = 7
        n = 3
        self.constants = [-1, -2.4, 0, 200, np.pi]
        sparsemat = scipy.sparse.rand(m=m, n=n, density=0.1, random_state=1)
        self.normal_params = [
            # Univariate
            (-1.0, 3.0),
            (1, 3),
            # Multivariate
            (np.random.uniform(size=10), np.eye(10)),
            (np.random.uniform(size=10), random_spd_matrix(10)),
            # Matrixvariate
            (
                np.random.uniform(size=(2, 2)),
                linops.SymmetricKronecker(
                    A=np.array([[1.0, 2.0], [2.0, 1.0]]),
                    B=np.array([[5.0, -1.0], [-1.0, 10.0]]),
                ).todense(),
            ),
            # Operatorvariate
            (
                np.array([1.0, -5.0]),
                linops.Matrix(A=np.array([[2.0, 1.0], [1.0, -0.1]])),
            ),
            (
                linops.Matrix(A=np.array([[0.0, -5.0]])),
                linops.Identity(shape=(2, 2)),
            ),
            (
                np.array([[1.0, 2.0], [-3.0, -0.4], [4.0, 1.0]]),
                linops.Kronecker(A=np.eye(3), B=5 * np.eye(2)),
            ),
            (
                linops.Matrix(A=sparsemat.todense()),
                linops.Kronecker(0.1 * linops.Identity(m), linops.Identity(n)),
            ),
            (
                linops.Matrix(A=np.random.uniform(size=(2, 2))),
                linops.SymmetricKronecker(
                    A=np.array([[1.0, 2.0], [2.0, 1.0]]),
                    B=np.array([[5.0, -1.0], [-1.0, 10.0]]),
                ),
            ),
            # Symmetric Kronecker Identical Factors
            (
                linops.Identity(shape=25),
                linops.SymmetricKronecker(A=linops.Identity(25)),
            ),
        ]

    def test_correct_instantiation(self):
        """Test whether different variants of the normal distribution are instances of
        Normal."""
        for mean, cov in self.normal_params:
            with self.subTest():
                dist = randvars.Normal(mean=mean, cov=cov)
                self.assertIsInstance(dist, randvars.Normal)

    def test_scalarmult(self):
        """Multiply a rv with a normal distribution with a scalar."""
        for (mean, cov), const in list(
            itertools.product(self.normal_params, self.constants)
        ):
            with self.subTest():
                normrv = const * randvars.Normal(mean=mean, cov=cov)

                self.assertIsInstance(normrv, probnum.RandomVariable)

                if const != 0:
                    self.assertIsInstance(normrv, randvars.Normal)
                else:
                    self.assertIsInstance(normrv, randvars.Constant)

    def test_addition_normal(self):
        """Add two random variables with a normal distribution."""
        for (mean0, cov0), (mean1, cov1) in list(
            itertools.product(self.normal_params, self.normal_params)
        ):
            with self.subTest():
                normrv0 = randvars.Normal(mean=mean0, cov=cov0)
                normrv1 = randvars.Normal(mean=mean1, cov=cov1)

                if normrv0.shape == normrv1.shape:
                    try:
                        newmean = mean0 + mean1
                        newcov = cov0 + cov1
                    except TypeError:
                        continue

                    self.assertIsInstance(normrv0 + normrv1, randvars.Normal)
                else:
                    with self.assertRaises(ValueError):
                        normrv_added = normrv0 + normrv1

    def test_rv_linop_kroneckercov(self):
        """Create a rv with a normal distribution with linear operator mean and
        Kronecker product kernels."""

        @linops.LinearOperator.broadcast_matvec
        def _matmul(v):
            return np.array([2 * v[0], 3 * v[1]])

        A = linops.LinearOperator(shape=(2, 2), dtype=np.double, matmul=_matmul)
        V = linops.Kronecker(A, A)
        randvars.Normal(mean=A, cov=V)

    def test_normal_dimension_mismatch(self):
        """Instantiating a normal distribution with mismatched mean and kernels should
        result in a ValueError."""
        for mean, cov in [
            (0, np.array([1, 2])),
            (np.array([1, 2]), np.array([1, 0])),
            (np.array([[-1, 0], [2, 1]]), np.eye(3)),
        ]:
            with self.subTest():
                err_msg = "Mean and kernels mismatch in normal distribution did not raise a ValueError."
                with self.assertRaises(ValueError, msg=err_msg):
                    assert randvars.Normal(mean=mean, cov=cov)

    def test_normal_instantiation(self):
        """Instantiation of a normal distribution with mixed mean and cov type."""
        for mean, cov in self.normal_params:
            with self.subTest():
                randvars.Normal(mean=mean, cov=cov)

    def test_normal_pdf(self):
        """Evaluate pdf at random input."""
        for mean, cov in self.normal_params:
            with self.subTest():
                dist = randvars.Normal(mean=mean, cov=cov)
                pass

    def test_normal_cdf(self):
        """Evaluate cdf at random input."""
        pass

    def test_sample(self):
        """Draw samples and check all sample dimensions."""
        for mean, cov in self.normal_params:
            with self.subTest():
                # TODO: check dimension of each realization in rv_sample
                rv = randvars.Normal(mean=mean, cov=cov, random_state=1)
                rv_sample = rv.sample(size=5)
                if not np.isscalar(rv.mean):
                    self.assertEqual(
                        rv_sample.shape[-rv.ndim :],
                        mean.shape,
                        msg="Realization shape does not match mean shape.",
                    )

    def test_sample_zero_cov(self):
        """Draw sample from distribution with zero kernels and check whether it equals
        the mean."""
        for mean, cov in self.normal_params:
            with self.subTest():
                rv = randvars.Normal(mean=mean, cov=0 * cov, random_state=1)
                rv_sample = rv.sample(size=1)
                assert_str = "Draw with kernels zero does not match mean."
                if isinstance(rv.mean, linops.LinearOperator):
                    self.assertAllClose(rv_sample, rv.mean.todense(), msg=assert_str)
                else:
                    self.assertAllClose(rv_sample, rv.mean, msg=assert_str)

    def test_symmetric_samples(self):
        """Samples from a normal distribution with symmetric Kronecker kernels of two
        symmetric matrices are symmetric."""
        np.random.seed(42)
        n = 3
        A = np.random.uniform(size=(n, n))
        A = 0.5 * (A + A.T) + n * np.eye(n)
        rv = randvars.Normal(
            mean=np.eye(A.shape[0]),
            cov=linops.SymmetricKronecker(A=A),
            random_state=1,
        )
        rv = rv.sample(size=10)
        for i, B in enumerate(rv):
            self.assertAllClose(
                B,
                B.T,
                atol=1e-5,
                rtol=1e-5,
                msg="Sample {} from symmetric Kronecker distribution is not symmetric.".format(
                    i
                ),
            )

    def test_indexing(self):
        """Indexing with Python integers yields a univariate normal distribution."""
        for mean, cov in self.normal_params:
            rv = randvars.Normal(mean=mean, cov=cov)

            with self.subTest():
                # Sample random index
                idx = tuple(np.random.randint(dim_size) for dim_size in rv.shape)

                # Index into distribution
                indexed_rv = rv[idx]

                self.assertIsInstance(indexed_rv, randvars.Normal)

                # Compare with expected parameter values
                if rv.ndim == 0:
                    flat_idx = ()
                elif rv.ndim == 1:
                    flat_idx = (idx[0],)
                else:
                    assert rv.ndim == 2

                    flat_idx = (idx[0] * rv.shape[1] + idx[1],)

                self.assertEqual(indexed_rv.shape, ())
                self.assertEqual(indexed_rv.mean, rv.dense_mean[idx])
                self.assertEqual(indexed_rv.var, rv.var[idx])
                self.assertEqual(indexed_rv.cov, rv.dense_cov[flat_idx + flat_idx])

    def test_slicing(self):
        """Slicing into a normal distribution yields a normal distribution of the same
        type."""
        for mean, cov in self.normal_params:
            rv = randvars.Normal(mean=mean, cov=cov)

            def _random_slice(dim_size):
                start = np.random.randint(0, dim_size)
                stop = np.random.randint(start + 1, dim_size + 1)

                return slice(start, stop)

            with self.subTest():
                # Sample random slice objects for each dimension
                slices = tuple(_random_slice(dim_size) for dim_size in rv.shape)

                # Get slice from distribution
                sliced_rv = rv[slices]

                # Compare with expected parameter values
                slice_mask = np.zeros_like(rv.dense_mean, dtype=np.bool_)
                slice_mask[slices] = True
                slice_mask = slice_mask.ravel()

                self.assertArrayEqual(sliced_rv.mean, rv.dense_mean[slices])
                self.assertArrayEqual(sliced_rv.var, rv.var[slices])

                if rv.ndim > 0:
                    self.assertArrayEqual(
                        sliced_rv.cov, rv.dense_cov[np.ix_(slice_mask, slice_mask)]
                    )
                else:
                    self.assertArrayEqual(sliced_rv.cov, rv.cov)

    def test_array_indexing(self):
        """Indexing with 1-dim integer arrays yields a multivariate normal."""
        for mean, cov in self.normal_params:
            rv = randvars.Normal(mean=mean, cov=cov)

            if rv.ndim == 0:
                continue

            with self.subTest():
                # Sample random indices
                idcs = tuple(
                    np.random.randint(dim_shape, size=10) for dim_shape in mean.shape
                )

                # Index into distribution
                indexed_rv = rv[idcs]

                self.assertIsInstance(indexed_rv, randvars.Normal)

                # Compare with expected parameter values
                if len(rv.shape) == 1:
                    flat_idcs = idcs[0]
                else:
                    assert len(rv.shape) == 2

                    flat_idcs = idcs[0] * rv.shape[1] + idcs[1]

                self.assertEqual(indexed_rv.shape, (10,))

                self.assertArrayEqual(indexed_rv.mean, rv.dense_mean[idcs])
                self.assertArrayEqual(indexed_rv.var, rv.var[idcs])
                self.assertArrayEqual(
                    indexed_rv.cov, rv.dense_cov[np.ix_(flat_idcs, flat_idcs)]
                )

    def test_array_indexing_broadcast(self):
        """Indexing with broadcasted integer arrays yields a matrixvariate normal."""
        for mean, cov in self.normal_params:
            rv = randvars.Normal(mean=mean, cov=cov)

            if rv.ndim != 2:
                continue

            with self.subTest():
                # Sample random indices
                idcs = np.ix_(
                    *tuple(
                        np.random.randint(dim_shape, size=10) for dim_shape in rv.shape
                    )
                )

                # Index into distribution
                indexed_rv = rv[idcs]

                self.assertIsInstance(indexed_rv, randvars.Normal)
                self.assertEqual(indexed_rv.shape, (10, 10))

                # Compare with expected parameter values
                flat_idcs = np.broadcast_arrays(*idcs)
                flat_idcs = flat_idcs[0] * rv.shape[1] + flat_idcs[1]
                flat_idcs = flat_idcs.ravel()

                self.assertArrayEqual(indexed_rv.mean, rv.dense_mean[idcs])
                self.assertArrayEqual(indexed_rv.var, rv.var[idcs])
                self.assertArrayEqual(
                    indexed_rv.cov, rv.dense_cov[np.ix_(flat_idcs, flat_idcs)]
                )

    def test_masking(self):
        """Masking a multivariate or matrixvariate normal yields a multivariate
        normal."""
        for mean, cov in self.normal_params:
            rv = randvars.Normal(mean=mean, cov=cov)

            with self.subTest():
                # Sample random indices
                idcs = tuple(
                    np.random.randint(dim_shape, size=10) for dim_shape in rv.shape
                )

                mask = np.zeros_like(rv.dense_mean, dtype=np.bool_)
                mask[idcs] = True

                # Mask distribution
                masked_rv = rv[mask]

                self.assertIsInstance(masked_rv, randvars.Normal)

                # Compare with expected parameter values
                flat_mask = mask.flatten()

                self.assertArrayEqual(masked_rv.mean, rv.dense_mean[mask])
                self.assertArrayEqual(masked_rv.var, rv.var[mask])

                if rv.ndim == 0:
                    self.assertArrayEqual(masked_rv.cov, rv.cov)
                else:
                    self.assertArrayEqual(
                        masked_rv.cov, rv.dense_cov[np.ix_(flat_mask, flat_mask)]
                    )


class UnivariateNormalTestCase(unittest.TestCase, NumpyAssertions):
    def setUp(self):
        self.params = (np.random.uniform(), np.random.gamma(shape=6, scale=1.0))

    def test_reshape_newaxis(self):
        dist = randvars.Normal(*self.params)

        for ndim in range(1, 3):
            for method in ["newaxis", "reshape"]:
                with self.subTest():
                    newshape = tuple([1] * ndim)

                    if method == "newaxis":
                        if ndim == 1:
                            newdist = dist[np.newaxis]
                        elif ndim == 2:
                            newdist = dist[np.newaxis, np.newaxis]
                    elif method == "reshape":
                        newdist = dist.reshape(newshape)

                    self.assertEqual(newdist.shape, newshape)
                    self.assertEqual(np.squeeze(newdist.mean), dist.mean)
                    self.assertEqual(np.squeeze(newdist.cov), dist.cov)

    def test_transpose(self):
        dist = randvars.Normal(*self.params)
        dist_t = dist.transpose()

        self.assertArrayEqual(dist_t.mean, dist.mean)
        self.assertArrayEqual(dist_t.cov, dist.cov)

        # Test sampling
        dist.random_state = 42
        dist_sample = dist.sample(size=5)

        dist_t.random_state = 42
        dist_t_sample = dist_t.sample(size=5)

        self.assertArrayEqual(dist_t_sample, dist_sample)

    def test_cov_cholesky_cov_cholesky_not_passed(self):
        """No cov_cholesky is passed in init.

        In this case, the "is_precomputed" flag is False, a cov_cholesky
        is computed on demand, but can also be computed manually with
        any damping factor.
        """
        mean, cov = self.params
        rv = randvars.Normal(mean, cov)

        with self.subTest("No Cholesky precomputed"):
            self.assertFalse(rv.cov_cholesky_is_precomputed)

        with self.subTest("Cholesky factor is computed correctly"):
            # The default damping factor 1e-12 does not mess up this test
            self.assertAllClose(rv.cov_cholesky, np.sqrt(rv.cov))

        with self.subTest("Cholesky is precomputed"):
            self.assertTrue(rv.cov_cholesky_is_precomputed)

    def test_precompute_cov_cholesky(self):
        mean, cov = self.params
        rv = randvars.Normal(mean, cov)

        with self.subTest("No Cholesky precomputed"):
            self.assertFalse(rv.cov_cholesky_is_precomputed)

        with self.subTest("Damping factor check"):

            rv.precompute_cov_cholesky(damping_factor=10.0)
            self.assertAllClose(rv.cov_cholesky, np.sqrt(rv.cov + 10.0))

        with self.subTest("Cholesky is precomputed"):
            self.assertTrue(rv.cov_cholesky_is_precomputed)

    def test_cov_cholesky_cov_cholesky_passed(self):
        """A value for cov_cholesky is passed in init.

        In this case, the "is_precomputed" flag is True, the
        cov_cholesky returns the argument that has been passed, but
        (p)recomputing overwrites the argument with a new factor.
        """
        mean, cov = self.params

        # This is purposely not the correct Cholesky factor for test reasons
        cov_cholesky = np.random.rand()

        rv = randvars.Normal(mean, cov, cov_cholesky=cov_cholesky)

        with self.subTest("Cholesky precomputed"):
            self.assertTrue(rv.cov_cholesky_is_precomputed)

        with self.subTest("Returns correct cov_cholesky"):
            self.assertAllClose(rv.cov_cholesky, cov_cholesky)

        with self.subTest("self.precompute raises exception"):
            with self.assertRaises(Exception):
                rv.precompute_cov_cholesky()


class MultivariateNormalTestCase(unittest.TestCase, NumpyAssertions):
    def setUp(self):
        self.params = (np.random.uniform(size=10), random_spd_matrix(10))

    def test_newaxis(self):
        vector_rv = randvars.Normal(*self.params)

        matrix_rv = vector_rv[:, np.newaxis]

        self.assertEqual(matrix_rv.shape, (10, 1))
        self.assertArrayEqual(np.squeeze(matrix_rv.mean), vector_rv.mean)
        self.assertArrayEqual(matrix_rv.cov, vector_rv.cov)

    def test_reshape(self):
        rv = randvars.Normal(*self.params)

        newshape = (5, 2)
        reshaped_rv = rv.reshape(newshape)

        self.assertArrayEqual(reshaped_rv.mean, rv.mean.reshape(newshape))
        self.assertArrayEqual(reshaped_rv.cov, rv.cov)

        # Test sampling
        rv.random_state = 42
        dist_sample = rv.sample(size=5)

        reshaped_rv.random_state = 42
        dist_reshape_sample = reshaped_rv.sample(size=5)

        self.assertArrayEqual(
            dist_reshape_sample, dist_sample.reshape((-1,) + newshape)
        )

    def test_transpose(self):
        rv = randvars.Normal(*self.params)
        transposed_rv = rv.transpose()

        self.assertArrayEqual(transposed_rv.mean, rv.mean)
        self.assertArrayEqual(transposed_rv.cov, rv.cov)

        # Test sampling
        rv.random_state = 42
        dist_sample = rv.sample(size=5)

        transposed_rv.random_state = 42
        dist_t_sample = transposed_rv.sample(size=5)

        self.assertArrayEqual(dist_t_sample, dist_sample)

    def test_cov_cholesky_cov_cholesky_not_passed(self):
        """No cov_cholesky is passed in init.

        In this case, the "is_precomputed" flag is False, a cov_cholesky
        is computed on demand, but can also be computed manually with
        any damping factor.
        """
        mean, cov = self.params

        rv = randvars.Normal(mean, cov)

        with self.subTest("No Cholesky precomputed"):
            self.assertFalse(rv.cov_cholesky_is_precomputed)

        with self.subTest("Cholesky factor is computed correctly"):
            # The default damping factor 1e-12 does not mess up this test
            self.assertAllClose(rv.cov_cholesky, np.linalg.cholesky(rv.cov))

        with self.subTest("Cholesky is precomputed"):
            self.assertTrue(rv.cov_cholesky_is_precomputed)

    def test_precompute_cov_cholesky(self):
        mean, cov = self.params
        rv = randvars.Normal(mean, cov)

        with self.subTest("No Cholesky precomputed"):
            self.assertFalse(rv.cov_cholesky_is_precomputed)

        with self.subTest("Damping factor check"):
            rv.precompute_cov_cholesky(damping_factor=10.0)
            self.assertAllClose(
                rv.cov_cholesky, np.linalg.cholesky(rv.cov + 10.0 * np.eye(len(rv.cov)))
            )

        with self.subTest("Cholesky is precomputed"):
            self.assertTrue(rv.cov_cholesky_is_precomputed)

    def test_cov_cholesky_cov_cholesky_passed(self):
        """A value for cov_cholesky is passed in init.

        In this case, the "is_precomputed" flag is True, the
        cov_cholesky returns the argument that has been passed, but
        (p)recomputing overwrites the argument with a new factor.
        """
        mean, cov = self.params

        # This is purposely not the correct Cholesky factor for test reasons
        cov_cholesky = np.random.rand(*cov.shape)

        rv = randvars.Normal(mean, cov, cov_cholesky=cov_cholesky)

        with self.subTest("Cholesky precomputed"):
            self.assertTrue(rv.cov_cholesky_is_precomputed)

        with self.subTest("Returns correct cov_cholesky"):
            self.assertAllClose(rv.cov_cholesky, cov_cholesky)

        with self.subTest("self.precompute raises exception"):
            with self.assertRaises(Exception):
                rv.precompute_cov_cholesky()

    def test_cholesky_cov_incompatible_types(self):
        """Test the behaviour of Normal.__init__ in the setup where the type of the
        Cholesky factor and the type of the covariance do not match."""
        mean, cov = self.params
        cov_cholesky = np.linalg.cholesky(cov)
        cov_cholesky_wrong_type = cov_cholesky.tolist()
        with self.subTest("Different type raises ValueError"):
            with self.assertRaises(TypeError):
                randvars.Normal(mean, cov, cov_cholesky=cov_cholesky_wrong_type)

        cov_cholesky_wrong_shape = cov_cholesky[1:]
        with self.subTest("Different shape raises ValueError"):
            with self.assertRaises(ValueError):
                randvars.Normal(mean, cov, cov_cholesky=cov_cholesky_wrong_shape)

        cov_cholesky_wrong_dtype = cov_cholesky.astype(int)
        with self.subTest("Different data type is promoted"):

            # Sanity check
            self.assertNotEqual(cov.dtype, cov_cholesky_wrong_dtype.dtype)

            # Assert data type of cov_cholesky is changed during __init__
            normal_new_dtype = randvars.Normal(
                mean, cov, cov_cholesky=cov_cholesky_wrong_dtype
            )
            self.assertEqual(
                normal_new_dtype.cov.dtype, normal_new_dtype.cov_cholesky.dtype
            )


class MatrixvariateNormalTestCase(unittest.TestCase, NumpyAssertions):
    def test_reshape(self):
        rv = randvars.Normal(
            mean=np.random.uniform(size=(4, 3)),
            cov=linops.Kronecker(
                A=random_spd_matrix(4), B=random_spd_matrix(3)
            ).todense(),
        )

        newshape = (2, 6)
        reshaped_rv = rv.reshape(newshape)

        self.assertArrayEqual(reshaped_rv.mean, rv.mean.reshape(newshape))
        self.assertArrayEqual(reshaped_rv.cov, rv.cov)

        # Test sampling
        rv.random_state = 42
        dist_sample = rv.sample(size=5)

        reshaped_rv.random_state = 42
        dist_reshape_sample = reshaped_rv.sample(size=5)

        self.assertArrayEqual(
            dist_reshape_sample, dist_sample.reshape((-1,) + newshape)
        )

    def test_transpose(self):
        rv = randvars.Normal(
            mean=np.random.uniform(size=(2, 2)), cov=random_spd_matrix(4)
        )
        transposed_rv = rv.transpose()

        self.assertArrayEqual(transposed_rv.mean, rv.mean.T)

        # Test covariance
        for ii, ij in itertools.product(range(2), range(2)):
            for ji, jj in itertools.product(range(2), range(2)):
                idx = (2 * ii + ij, 2 * ji + jj)
                idx_t = (2 * ij + ii, 2 * jj + ji)

                self.assertEqual(transposed_rv.cov[idx_t], rv.cov[idx])

        # Sadly, sampling is not stable w.r.t. permutations of variables

    def test_cov_cholesky_cov_cholesky_not_passed(self):
        """No cov_cholesky is passed in init.

        In this case, the "is_precomputed" flag is False, a cov_cholesky
        is computed on demand, but can also be computed manually with
        any damping factor.
        """
        rv = randvars.Normal(
            mean=np.random.uniform(size=(2, 2)), cov=random_spd_matrix(4)
        )

        with self.subTest("No Cholesky precomputed"):
            self.assertFalse(rv.cov_cholesky_is_precomputed)

        with self.subTest("Cholesky factor is computed correctly"):
            # The default damping factor 1e-12 does not mess up this test
            self.assertAllClose(rv.cov_cholesky, np.linalg.cholesky(rv.cov))

        with self.subTest("Cholesky is precomputed"):
            self.assertTrue(rv.cov_cholesky_is_precomputed)

    def test_precompute_cov_cholesky(self):
        rv = randvars.Normal(
            mean=np.random.uniform(size=(2, 2)), cov=random_spd_matrix(4)
        )

        with self.subTest("No Cholesky precomputed"):
            self.assertFalse(rv.cov_cholesky_is_precomputed)

        with self.subTest("Damping factor check"):
            rv.precompute_cov_cholesky(damping_factor=10.0)
            self.assertAllClose(
                rv.cov_cholesky, np.linalg.cholesky(rv.cov + 10.0 * np.eye(len(rv.cov)))
            )

        with self.subTest("Cholesky is precomputed"):
            self.assertTrue(rv.cov_cholesky_is_precomputed)

    def test_cov_cholesky_cov_cholesky_passed(self):
        """A value for cov_cholesky is passed in init.

        In this case, the "is_precomputed" flag is True, the
        cov_cholesky returns the argument that has been passed, but
        (p)recomputing overwrites the argument with a new factor.
        """
        # This is purposely not the correct Cholesky factor for test reasons
        cov_cholesky = np.random.rand(4, 4)

        rv = randvars.Normal(
            mean=np.random.uniform(size=(2, 2)),
            cov=random_spd_matrix(4),
            cov_cholesky=cov_cholesky,
        )

        with self.subTest("Cholesky precomputed"):
            self.assertTrue(rv.cov_cholesky_is_precomputed)

        with self.subTest("Returns correct cov_cholesky"):
            self.assertAllClose(rv.cov_cholesky, cov_cholesky)

        with self.subTest("self.precompute raises exception"):
            with self.assertRaises(Exception):
                rv.precompute_cov_cholesky()


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from probnum.utils import randomutils
from tests.testing import NumpyAssertions


class RandomUtilsTestCase(unittest.TestCase, NumpyAssertions):
    """Test case for utility functions handling objects dealing with randomness."""

    def setUp(self) -> None:
        self.seed = 42
        self.random_state = np.random.RandomState(seed=self.seed)
        self.random_generator = np.random.default_rng(seed=self.seed)
        self.random_generator_list = [np.random.default_rng(seed=s) for s in range(5)]

    def test_derive_random_seed_invariant_random_state(self):
        """Test whether deriving a random seed leaves the original random states
        untouched."""
        # Original random states
        rs_state = self.random_state.get_state()[1]
        rng_state = self.random_generator.bit_generator.state["state"]["state"]
        rng_list_states = [
            rng.bit_generator.state["state"]["state"]
            for rng in self.random_generator_list
        ]

        # Combine RandomState and Generator object
        _ = randomutils.derive_random_seed(self.random_state, self.random_generator)
        self.assertArrayEqual(rs_state, self.random_state.get_state()[1])
        self.assertEqual(
            rng_state, self.random_generator.bit_generator.state["state"]["state"]
        )

        # Combine list of generators
        _ = randomutils.derive_random_seed(*self.random_generator_list)
        self.assertArrayEqual(
            rng_list_states,
            [
                rng.bit_generator.state["state"]["state"]
                for rng in self.random_generator_list
            ],
        )



"""Make a ProbNum ODE solution out of a scipy ODE solution."""
import numpy as np
from scipy.integrate._ivp.common import OdeSolution

from probnum import _randomvariablelist, diffeq, randvars
from probnum.filtsmooth.timeseriesposterior import (
    DenseOutputLocationArgType,
    DenseOutputValueType,
)


class WrappedScipyODESolution(diffeq.ODESolution):
    """Make a ProbNum ODESolution out of a SciPy OdeSolution."""

    def __init__(self, scipy_solution: OdeSolution, rvs: list):
        self.scipy_solution = scipy_solution

        # rvs is of the type `list` of `RandomVariable` and can therefore be
        # directly transformed into a _RandomVariableList
        rv_states = _randomvariablelist._RandomVariableList(rvs)
        super().__init__(locations=scipy_solution.ts, states=rv_states)

    def __call__(self, t: DenseOutputLocationArgType) -> DenseOutputValueType:
        """Evaluate the time-continuous solution at time t.

        Parameters
        ----------
        t
        Location / time at which to evaluate the continuous ODE solution.

        Returns
        -------
        randvars.RandomVariable or _randomvariablelist._RandomVariableList
            Estimate of the states at time ``t`` based on a fourth order polynomial.
        """

        states = self.scipy_solution(t).T
        if np.isscalar(t):
            solution_as_rv = randvars.Constant(states)
        else:
            solution_as_rv = _randomvariablelist._RandomVariableList(
                [randvars.Constant(state) for state in states]
            )
        return solution_as_rv



import numpy as np
import pytest

import probnum.statespace as pnss


@pytest.fixture
def precon():
    some_order = 3
    some_dim = 1
    return pnss.NordsieckLikeCoordinates.from_order(some_order, some_dim)


def test_call(precon):
    P = precon(0.5)
    assert isinstance(P, np.ndarray)


def test_inverse(precon):
    P, Pinv = precon(0.5), precon.inverse(0.5)
    np.testing.assert_allclose(P @ Pinv, np.eye(*P.shape))
    np.testing.assert_allclose(Pinv @ P, np.eye(*P.shape))



"""Tests for functions generating random spd matrices."""

import unittest

import numpy as np

from probnum.problems.zoo.linalg import random_sparse_spd_matrix, random_spd_matrix
from tests.testing import NumpyAssertions


class RandomSPDMatrixTestCase(unittest.TestCase, NumpyAssertions):
    """Test case for random spd matrices as test problems."""

    def setUp(self) -> None:
        """Define parameters and define test problems."""
        self.rng = np.random.default_rng(42)
        self.dim_list = [1, 2, 25, 100, 250]
        self.spd_matrices = [
            random_spd_matrix(dim=n, random_state=self.rng) for n in self.dim_list
        ]
        self.density = 0.01
        self.sparse_spd_matrices = [
            random_sparse_spd_matrix(dim=n, density=self.density, random_state=self.rng)
            for n in self.dim_list
        ]
        self.matrices = self.spd_matrices + self.sparse_spd_matrices

    def test_returns_ndarray(self):
        """Test whether test problems return ndarrays."""
        for mat in self.matrices:
            with self.subTest():
                self.assertIsInstance(mat, np.ndarray)

    def test_dimension(self):
        """Test whether matrix dimension matches specified dimension."""
        for mat, dim in zip(self.matrices, self.dim_list + self.dim_list):
            with self.subTest():
                self.assertEqual(
                    mat.shape[0], dim, msg="Matrix dimension does not match argument."
                )

    def test_symmetric(self):
        """Test whether the matrix is symmetric."""
        for mat in self.matrices:
            with self.subTest():
                self.assertArrayEqual(mat, mat.T, msg="Matrix is not symmetric.")

    def test_positive_definite(self):
        """Test whether the matrix is positive definite."""
        for mat in self.matrices:
            eigvals = np.linalg.eigvals(mat)
            with self.subTest():
                self.assertTrue(
                    np.all(eigvals > 0.0), msg="Eigenvalues are not all positive."
                )

    def test_spectrum_matches_given(self):
        """Test whether the spectrum of the test problem matches the provided
        spectrum."""
        dim = 10
        spectrum = np.sort(self.rng.uniform(0.1, 1, size=dim))
        spdmat = random_spd_matrix(dim=dim, spectrum=spectrum, random_state=self.rng)
        eigvals = np.sort(np.linalg.eigvals(spdmat))
        self.assertAllClose(
            spectrum,
            eigvals,
            msg="Provided spectrum doesn't match actual.",
        )

    def test_negative_eigenvalues_throws_error(self):
        """Test whether a non-positive spectrum throws an error."""
        with self.assertRaises(ValueError):
            random_spd_matrix(dim=3, spectrum=[-1, 1, 2], random_state=self.rng)

    def test_matrix_is_sparse(self):
        """Test whether the matrix has a sufficient degree of sparsity."""
        for sparsemat in self.sparse_spd_matrices:
            with self.subTest():
                emp_density = (
                    np.sum(sparsemat != 0.0) - sparsemat.shape[0]
                ) / sparsemat.shape[0] ** 2
                self.assertLess(
                    emp_density,
                    self.density * 2,
                    msg=f"Matrix has {emp_density}n "
                    f"non-zero entries, which doesnt match the "
                    f"given degree of sparsity.",
                )



import pytest

import probnum.problems.zoo.diffeq as diffeq_zoo

# Jax dependency handling
# pylint: disable=unused-import
try:
    import jax
    import jax.numpy as jnp
    from jax.config import config

    config.update("jax_enable_x64", True)

    JAX_AVAILABLE = True

    IVPs = [diffeq_zoo.threebody_jax(), diffeq_zoo.vanderpol_jax()]


except ImportError:
    JAX_AVAILABLE = False
    IVPs = []


# Pytest decorators to select tests for each case
only_if_jax_available = pytest.mark.skipif(not JAX_AVAILABLE, reason="requires jax")
only_if_jax_is_not_available = pytest.mark.skipif(
    JAX_AVAILABLE,
    reason="Imports will be successful, thus catching the ImportError will fail",
)

# Tests for when JAX is available


@pytest.mark.parametrize("ivp_jax", IVPs)
@only_if_jax_available
def test_f(ivp_jax):
    ivp_jax.f(ivp_jax.t0, ivp_jax.y0)


@pytest.mark.parametrize("ivp_jax", IVPs)
@only_if_jax_available
def test_df(ivp_jax):
    ivp_jax.df(ivp_jax.t0, ivp_jax.y0)


@pytest.mark.parametrize("ivp_jax", IVPs)
@only_if_jax_available
def test_ddf(ivp_jax):
    ivp_jax.ddf(ivp_jax.t0, ivp_jax.y0)


# Tests for when JAX is not available


@only_if_jax_is_not_available
def test_threebody():
    with pytest.raises(ImportError):
        diffeq_zoo.threebody_jax()


@only_if_jax_is_not_available
def test_vanderpol():
    with pytest.raises(ImportError):
        diffeq_zoo.vanderpol_jax()



"""Update the belief over the parameters with observations in a 1D quadratic
optimization problem."""

from typing import Union

import numpy as np

import probnum as pn
from probnum import linops, randvars
from probnum.type import FloatArgType


def gaussian_belief_update(
    fun_params0: pn.RandomVariable,
    action: FloatArgType,
    observation: FloatArgType,
    noise_cov: Union[np.ndarray, linops.LinearOperator],
) -> randvars.RandomVariable:
    """Update the belief over the parameters with an observation.

    Parameters
    ----------
    fun_params0 :
        Belief over the parameters of the quadratic objective.
    action :
        Action of the probabilistic quadratic optimizer.
    observation :
        Observation of the problem corresponding to the given `action`.
    noise_cov :
        *shape=(3, 3)* -- Covariance of the noise on the parameters of the quadratic
        objective given by the assumed observation model.
    """
    # Feature vector
    x = np.asarray(action).reshape(1, -1)
    Phi = np.vstack((0.5 * x ** 2, x, np.ones_like(x)))

    # Mean and covariance
    mu = fun_params0.mean
    Sigma = fun_params0.cov

    # Gram matrix
    gram = Phi.T @ (Sigma + noise_cov) @ Phi

    # Posterior Mean
    m = mu + Sigma @ Phi @ np.linalg.solve(gram, observation - Phi.T @ mu)

    # Posterior Covariance
    S = Sigma - Sigma @ Phi @ np.linalg.solve(gram, Phi.T @ Sigma)

    return randvars.Normal(mean=m, cov=S)



"""Convenience functions for Gaussian filtering and smoothing.

References
----------
.. [1] https://arxiv.org/pdf/1610.05261.pdf
.. [2] https://arxiv.org/abs/1807.09737
.. [3] https://arxiv.org/abs/1810.03440
.. [4] https://arxiv.org/pdf/2004.00623.pdf
"""

import numpy as np

from probnum import randvars, statespace

from .. import steprule
from ..ode import IVP
from .ivpfiltsmooth import GaussianIVPFilter

__all__ = ["probsolve_ivp"]


def probsolve_ivp(
    f,
    t0,
    tmax,
    y0,
    df=None,
    method="EK0",
    dense_output=True,
    algo_order=2,
    adaptive=True,
    atol=1e-2,
    rtol=1e-2,
    step=None,
):
    r"""Solve initial value problem with Gaussian filtering and smoothing.

    Numerically computes a Gauss-Markov process which solves numerically
    the initial value problem (IVP) based on a system of first order
    ordinary differential equations (ODEs)

    .. math:: \\dot x(t) = f(t, x(t)), \\quad x(t_0) = x_0,
        \\quad t \\in [t_0, T]

    by regarding it as a (nonlinear) Gaussian filtering (and smoothing)
    problem [3]_. For some configurations it recovers certain multistep
    methods [1]_.
    Convergence rates of filtering [2]_ and smoothing [4]_ are
    comparable to those of methods of Runge-Kutta type.


    This function turns a prior-string into an :class:`ODEPrior`, a
    method-string into a filter/smoother of class :class:`GaussFiltSmooth`, creates a
    :class:`GaussianIVPFilter` object and calls the :meth:`solve()` method. For
    advanced usage we recommend to do this process manually which
    enables advanced methods of tuning the algorithm.

    This function supports the methods:
    extended Kalman filtering based on a zero-th order Taylor
    approximation (EKF0),
    extended Kalman filtering (EKF1),
    unscented Kalman filtering (UKF),
    extended Kalman smoothing based on a zero-th order Taylor
    approximation (EKS0),
    extended Kalman smoothing (EKS1), and
    unscented Kalman smoothing (UKS).

    For adaptive step-size selection of ODE filters, we implement the
    scheme proposed by Schober et al. (2019), and further examined by Bosch et al (2021),
    where the local error estimate is derived from the local, calibrated
    uncertainty estimate.

    Arguments
    ---------
    f :
        ODE vector field.
    t0 :
        Initial time point.
    tmax :
        Final time point.
    y0 :
        Initial value.
    df :
        Jacobian of the ODE vector field.
    adaptive :
        Whether to use adaptive steps or not. Default is `True`.
    atol : float
        Absolute tolerance  of the adaptive step-size selection scheme.
        Optional. Default is ``1e-4``.
    rtol : float
        Relative tolerance   of the adaptive step-size selection scheme.
        Optional. Default is ``1e-4``.
    step :
        Step size. If atol and rtol are not specified, this step-size is used for a fixed-step ODE solver.
        If they are specified, this only affects the first step. Optional.
        Default is None, in which case the first step is chosen as :math:`0.01 \cdot |y_0|/|f(t_0, y_0)|`.
    algo_order
        Order of the algorithm. This amounts to choosing the order of integration (``ordint``) of an integrated Brownian motion prior.
        For too high orders, process noise covariance matrices become singular. For IBM, this maximum seems to be :`q=11` (using standard ``float64``).
        It is possible that higher orders may work for you.
        The type of prior relates to prior assumptions about the
        derivative of the solution.
        The higher the order of the algorithm, the faster the convergence, but also, the higher-dimensional (and thus the costlier) the state space.
    method : str, optional
        Which method is to be used. Default is ``EK0`` which is the
        method proposed by Schober et al.. The available
        options are

        ================================================  ==============
         Extended Kalman filtering/smoothing (0th order)  ``'EK0'``
         Extended Kalman filtering/smoothing (1st order)  ``'EK1'``
        ================================================  ==============

        First order extended Kalman filtering and smoothing methods (``EK1``)
        require Jacobians of the RHS-vector field of the IVP.
        That is, the argument ``df`` needs to be specified.
        They are likely to perform better than zeroth order methods in
        terms of (A-)stability and "meaningful uncertainty estimates".
        While we recommend to use correct capitalization for the method string,
        lower-case letters will be capitalized internally.
    dense_output : bool
        Whether we want dense output. Optional. Default is ``True``. For the ODE filter,
        dense output requires smoothing, so if ``dense_output`` is False, no smoothing is performed;
        but when it is ``True``, the filter solution is smoothed.

    Returns
    -------
    solution : KalmanODESolution
        Solution of the ODE problem.

        Can be evaluated at and sampled from at arbitrary grid points.
        Further, it contains fields:

        t : :obj:`np.ndarray`, shape=(N,)
            Mesh used by the solver to compute the solution.
            It includes the initial time :math:`t_0` but not necessarily the
            final time :math:`T`.
        y : :obj:`list` of :obj:`RandomVariable`, length=N
            Discrete-time solution at times :math:`t_1, ..., t_N`,
            as a list of random variables.
            The means and covariances can be accessed with ``solution.y.mean``
            and ``solution.y.cov``.

    See Also
    --------
    GaussianIVPFilter : Solve IVPs with Gaussian filtering and smoothing
    KalmanODESolution : Solution of ODE problems based on Gaussian filtering and smoothing.

    References
    ----------
    .. [1] Schober, M., Särkkä, S. and Hennig, P..
        A probabilistic model for the numerical solution of initial
        value problems.
        Statistics and Computing, 2019.
    .. [2] Kersting, H., Sullivan, T.J., and Hennig, P..
        Convergence rates of Gaussian ODE filters.
        2019.
    .. [3] Tronarp, F., Kersting, H., Särkkä, S., and Hennig, P..
        Probabilistic solutions to ordinary differential equations as
        non-linear Bayesian filtering: a new perspective.
        Statistics and Computing, 2019.
    .. [4] Tronarp, F., Särkkä, S., and Hennig, P..
        Bayesian ODE solvers: the maximum a posteriori estimate.
        2019.


    Examples
    --------
    >>> from probnum.diffeq import logistic, probsolve_ivp
    >>> from probnum import randvars
    >>> import numpy as np

    Solve a simple logistic ODE with fixed steps.

    >>> def f(t, x):
    ...     return 4*x*(1-x)
    >>>
    >>> y0 = np.array([0.15])
    >>> t0, tmax = 0., 1.5
    >>> solution = probsolve_ivp(f, t0, tmax, y0, step=0.1, adaptive=False)
    >>> print(np.round(solution.states.mean, 2))
    [[0.15]
     [0.21]
     [0.28]
     [0.37]
     [0.47]
     [0.57]
     [0.66]
     [0.74]
     [0.81]
     [0.87]
     [0.91]
     [0.94]
     [0.96]
     [0.97]
     [0.98]
     [0.99]]


    Other methods are easily accessible.

    >>> def df(t, x):
    ...     return np.array([4. - 8 * x])
    >>> solution = probsolve_ivp(f, t0, tmax, y0, df=df, method="EK1", algo_order=2, step=0.1, adaptive=False)
    >>> print(np.round(solution.states.mean, 2))
    [[0.15]
     [0.21]
     [0.28]
     [0.37]
     [0.47]
     [0.57]
     [0.66]
     [0.74]
     [0.81]
     [0.87]
     [0.91]
     [0.93]
     [0.96]
     [0.97]
     [0.98]
     [0.99]]

    """

    # Create IVP object
    ivp = IVP(
        timespan=(t0, tmax),
        initrv=randvars.Constant(np.asarray(y0)),
        rhs=f,
        jac=df,
    )

    # Create steprule
    if adaptive is True:
        if atol is None or rtol is None:
            raise ValueError(
                "Please provide absolute and relative tolerance for adaptive steps."
            )
        firststep = step if step is not None else steprule.propose_firststep(ivp)
        stprl = steprule.AdaptiveSteps(firststep=firststep, atol=atol, rtol=rtol)
    else:
        stprl = steprule.ConstantSteps(step)

    # Create solver
    prior = statespace.IBM(
        ordint=algo_order,
        spatialdim=ivp.dimension,
        forward_implementation="sqrt",
        backward_implementation="sqrt",
    )

    if method.upper() not in ["EK0", "EK1"]:
        raise ValueError("Method is not supported.")
    measmod = GaussianIVPFilter.string_to_measurement_model(method, ivp, prior)
    solver = GaussianIVPFilter.construct_with_rk_init(
        ivp, prior, measmod, with_smoothing=dense_output
    )

    return solver.solve(steprule=stprl)



"""Test cases for Bayesian quadrature."""

import numpy as np
import pytest
from scipy.integrate import quad

from probnum.quad import BayesianQuadrature, sample_from_measure
from probnum.randvars import Normal


@pytest.mark.parametrize("input_dim", [1], ids=["dim1"])
def test_type_1d(f1d, kernel, measure):
    """Test that BQ outputs normal random variables for 1D integrands."""
    # pylint: disable=invalid-name
    bq = BayesianQuadrature(kernel=kernel, policy=sample_from_measure)
    integral, _ = bq.integrate(fun=f1d, measure=measure, nevals=10)
    assert isinstance(integral, Normal)


@pytest.mark.parametrize("input_dim", [1])
def test_integral_values_1d(f1d, kernel, measure):
    """Test numerically that BQ computes 1D integrals correctly."""

    # numerical integral
    # pylint: disable=invalid-name
    def integrand(x):
        return f1d(x) * measure(x)

    # pylint: disable=invalid-name
    bq = BayesianQuadrature(kernel=kernel, policy=sample_from_measure)
    num_integral, _ = quad(integrand, measure.domain[0], measure.domain[1])
    bq_integral, _ = bq.integrate(fun=f1d, measure=measure, nevals=250)
    np.testing.assert_almost_equal(bq_integral.mean, num_integral, decimal=2)



"""Tests for random variable arithmetic."""

import numpy as np
import pytest

from probnum import randvars, utils


@pytest.fixture
def constant():
    return randvars.Constant(support=np.arange(10, 12))


@pytest.fixture
def normal(cov_cholesky):
    return randvars.Normal(
        mean=np.arange(4, 6), cov=np.diag(np.arange(5, 7)), cov_cholesky=cov_cholesky
    )


@pytest.mark.parametrize("cov_cholesky", [None, np.diag(np.sqrt(np.arange(5, 7)))])
def test_constant_normal_sum_left(constant, normal):
    """Assert that mean and covariance follow the correct formula and that a Cholesky
    factor is preserved if it existed before."""
    sum_of_rvs = constant + normal

    np.testing.assert_allclose(sum_of_rvs.mean, constant.support + normal.mean)
    np.testing.assert_allclose(sum_of_rvs.cov, normal.cov)
    assert sum_of_rvs.cov_cholesky_is_precomputed == normal.cov_cholesky_is_precomputed
    if sum_of_rvs.cov_cholesky_is_precomputed:
        np.testing.assert_allclose(sum_of_rvs.cov_cholesky, normal.cov_cholesky)


@pytest.mark.parametrize("cov_cholesky", [None, np.diag(np.sqrt(np.arange(5, 7)))])
def test_constant_normal_sum_right(constant, normal):
    """Assert that mean and covariance follow the correct formula and that a Cholesky
    factor is preserved if it existed before."""
    sum_of_rvs = normal + constant
    np.testing.assert_allclose(sum_of_rvs.mean, normal.mean + constant.support)
    np.testing.assert_allclose(sum_of_rvs.cov, normal.cov)
    assert sum_of_rvs.cov_cholesky_is_precomputed == normal.cov_cholesky_is_precomputed
    if sum_of_rvs.cov_cholesky_is_precomputed:
        np.testing.assert_allclose(sum_of_rvs.cov_cholesky, normal.cov_cholesky)


@pytest.mark.parametrize("cov_cholesky", [None, np.diag(np.sqrt(np.arange(5, 7)))])
def test_constant_normal_subtraction_left(constant, normal):
    """Assert that mean and covariance follow the correct formula and that a Cholesky
    factor is preserved if it existed before."""
    diff_of_rvs = constant - normal

    np.testing.assert_allclose(diff_of_rvs.mean, constant.support - normal.mean)
    np.testing.assert_allclose(diff_of_rvs.cov, normal.cov)
    assert diff_of_rvs.cov_cholesky_is_precomputed == normal.cov_cholesky_is_precomputed
    if diff_of_rvs.cov_cholesky_is_precomputed:
        np.testing.assert_allclose(diff_of_rvs.cov_cholesky, normal.cov_cholesky)


@pytest.mark.parametrize("cov_cholesky", [None, np.diag(np.sqrt(np.arange(5, 7)))])
def test_constant_normal_subtraction_right(constant, normal):
    """Assert that mean and covariance follow the correct formula and that a Cholesky
    factor is preserved if it existed before."""
    diff_of_rvs = normal - constant
    np.testing.assert_allclose(diff_of_rvs.mean, normal.mean - constant.support)
    np.testing.assert_allclose(diff_of_rvs.cov, normal.cov)
    assert diff_of_rvs.cov_cholesky_is_precomputed == normal.cov_cholesky_is_precomputed
    if diff_of_rvs.cov_cholesky_is_precomputed:
        np.testing.assert_allclose(diff_of_rvs.cov_cholesky, normal.cov_cholesky)


@pytest.mark.parametrize("cov_cholesky", [None, np.diag(np.sqrt(np.arange(5, 7)))])
def test_constant_normal_matrix_multiplication_right(constant, normal):
    """Assert that mean and covariance follow the correct formula and that a Cholesky
    factor is preserved if it existed before."""
    matrix_product = constant @ normal
    np.testing.assert_allclose(matrix_product.mean, constant.support @ normal.mean)
    np.testing.assert_allclose(
        matrix_product.cov, constant.support @ normal.cov @ constant.support.T
    )
    assert (
        matrix_product.cov_cholesky_is_precomputed == normal.cov_cholesky_is_precomputed
    )
    if matrix_product.cov_cholesky_is_precomputed:
        np.testing.assert_allclose(
            matrix_product.cov_cholesky,
            utils.linalg.cholesky_update(constant.support @ normal.cov_cholesky),
        )


@pytest.mark.parametrize("cov_cholesky", [None, np.diag(np.sqrt(np.arange(5, 7)))])
def test_constant_normal_matrix_multiplication_left(constant, normal):
    """Assert that mean and covariance follow the correct formula and that a Cholesky
    factor is preserved if it existed before."""
    matrix_product = normal @ constant
    np.testing.assert_allclose(matrix_product.mean, normal.mean @ constant.support)
    np.testing.assert_allclose(
        matrix_product.cov, constant.support.T @ normal.cov @ constant.support
    )
    assert (
        matrix_product.cov_cholesky_is_precomputed == normal.cov_cholesky_is_precomputed
    )
    if matrix_product.cov_cholesky_is_precomputed:
        np.testing.assert_allclose(
            matrix_product.cov_cholesky,
            utils.linalg.cholesky_update(constant.support.T @ normal.cov_cholesky),
        )


@pytest.mark.parametrize("cov_cholesky", [None, np.diag(np.sqrt(np.arange(5, 7)))])
def test_constant_normal_multiplication_right(constant, normal):
    """Assert that mean and covariance follow the correct formula and that a Cholesky
    factor is preserved if it existed before."""

    # Elementwise multiplication only supported for Constant RVs of size 1.
    scalar_like_constant = constant[0]
    element_product = scalar_like_constant * normal

    np.testing.assert_allclose(
        element_product.mean, scalar_like_constant.support * normal.mean
    )
    np.testing.assert_allclose(
        element_product.cov,
        scalar_like_constant.support ** 2 * normal.cov,
    )
    assert (
        element_product.cov_cholesky_is_precomputed
        == normal.cov_cholesky_is_precomputed
    )
    if element_product.cov_cholesky_is_precomputed:
        np.testing.assert_allclose(
            element_product.cov_cholesky,
            scalar_like_constant.support * normal.cov_cholesky,
        )


@pytest.mark.parametrize("cov_cholesky", [None, np.diag(np.sqrt(np.arange(5, 7)))])
def test_constant_normal_multiplication_left(constant, normal):
    """Assert that mean and covariance follow the correct formula and that a Cholesky
    factor is preserved if it existed before."""

    # Elementwise multiplication only supported for Constant RVs of size 1.
    scalar_like_constant = constant[0]
    element_product = normal * scalar_like_constant

    np.testing.assert_allclose(
        element_product.mean, normal.mean * scalar_like_constant.support
    )
    np.testing.assert_allclose(
        element_product.cov,
        normal.cov * scalar_like_constant.support ** 2,
    )
    assert (
        element_product.cov_cholesky_is_precomputed
        == normal.cov_cholesky_is_precomputed
    )
    if element_product.cov_cholesky_is_precomputed:
        np.testing.assert_allclose(
            element_product.cov_cholesky,
            normal.cov_cholesky * scalar_like_constant.support,
        )


@pytest.mark.parametrize("cov_cholesky", [None, np.diag(np.sqrt(np.arange(5, 7)))])
def test_constant_normal_division(constant, normal):
    """Assert that mean and covariance follow the correct formula and that a Cholesky
    factor is preserved if it existed before."""

    # Elementwise multiplication only supported for Constant RVs of size 1.
    scalar_like_constant = constant[0]
    element_product = normal / scalar_like_constant

    np.testing.assert_allclose(
        element_product.mean, normal.mean / scalar_like_constant.support
    )
    np.testing.assert_allclose(
        element_product.cov,
        normal.cov / scalar_like_constant.support ** 2,
    )
    assert (
        element_product.cov_cholesky_is_precomputed
        == normal.cov_cholesky_is_precomputed
    )
    if element_product.cov_cholesky_is_precomputed:
        np.testing.assert_allclose(
            element_product.cov_cholesky,
            normal.cov_cholesky / scalar_like_constant.support,
        )



"""Bayesian Filtering and Smoothing."""

from .bayesfiltsmooth import BayesFiltSmooth
from .gaussfiltsmooth import (
    ContinuousEKFComponent,
    ContinuousUKFComponent,
    DiscreteEKFComponent,
    DiscreteUKFComponent,
    EKFComponent,
    FilteringPosterior,
    IteratedDiscreteComponent,
    Kalman,
    KalmanPosterior,
    SmoothingPosterior,
    StoppingCriterion,
    UKFComponent,
    UnscentedTransform,
)
from .particlefiltsmooth import (
    ParticleFilter,
    ParticleFilterPosterior,
    effective_number_of_events,
)
from .timeseriesposterior import TimeSeriesPosterior

# Public classes and functions. Order is reflected in documentation.
__all__ = [
    "BayesFiltSmooth",
    "Kalman",
    "EKFComponent",
    "ContinuousEKFComponent",
    "DiscreteEKFComponent",
    "UKFComponent",
    "ContinuousUKFComponent",
    "DiscreteUKFComponent",
    "UnscentedTransform",
    "TimeSeriesPosterior",
    "KalmanPosterior",
    "FilteringPosterior",
    "SmoothingPosterior",
    "StoppingCriterion",
    "IteratedDiscreteComponent",
    "ParticleFilter",
    "ParticleFilterPosterior",
    "effective_number_of_events",
]



import numpy as np
import pytest

import probnum.statespace as pnss


@pytest.fixture
def dt():
    return 0.1


@pytest.fixture
def Ah_closedform(dt):
    return np.array([[1.0, dt], [0.0, 1.0]])


@pytest.fixture
def Qh_closedform(dt):
    return np.array([[dt ** 3 / 3, dt ** 2 / 2], [dt ** 2 / 2, dt]])


@pytest.fixture
def F():
    return np.array([[0.0, 1.0], [0.0, 0.0]])


# dispersion matrices in different shapes


def L0():
    return np.array([0.0, 1.0])


def L1():
    return np.array([[0.0], [1.0]])


@pytest.mark.parametrize("L", [L0(), L1()])
def test_matrix_fraction_decomposition(F, L, dt, Ah_closedform, Qh_closedform):
    """Test MFD against a closed-form IBM solution."""
    Ah, Qh, _ = pnss.matrix_fraction_decomposition(F, L, dt=dt)

    np.testing.assert_allclose(Ah, Ah_closedform)
    np.testing.assert_allclose(Qh, Qh_closedform)



import numpy as np

from probnum.problems import InitialValueProblem

__all__ = ["threebody", "vanderpol", "rigidbody"]


def threebody(t0=0.0, tmax=17.0652165601579625588917206249, y0=None):
    r"""Initial value problem (IVP) based on a three-body problem.

    Let the initial conditions be :math:`y = (y_1, y_2, \dot{y}_1, \dot{y}_2)^T`.
    This function implements the second-order three-body problem as a system of
    first-order ODEs, which is defined as follows: [1]_

    .. math::

        f(t, y) =
        \begin{pmatrix}
            \dot{y_1} \\
            \dot{y_2} \\
            y_1 + 2 \dot{y}_2 - \frac{(1 - \mu) (y_1 + \mu)}{d_1}
                - \frac{\mu (y_1 - (1 - \mu))}{d_2} \\
            y_2 - 2 \dot{y}_1 - \frac{(1 - \mu) y_2}{d_1} - \frac{\mu y_2}{d_2}
        \end{pmatrix}

    with

    .. math::

        d_1 &= ((y_1 + \mu)^2 + y_2^2)^{\frac{3}{2}} \\
        d_2 &= ((y_1 - (1 - \mu))^2 + y_2^2)^{\frac{3}{2}}

    and a constant parameter  :math:`\mu = 0.012277471` denoting the standardized moon mass.

    Parameters
    ----------
    t0
        Initial time. Default is ``0.0``.
    tmax
        Final time. Default is ``17.0652165601579625588917206249`` which is the period of the solution.
    y0
        *(shape=(4, ))* -- Initial value. Default is ``[0.994, 0, 0, -2.00158510637908252240537862224]``.

    Returns
    -------
    InitialValueProblem
        InitialValueProblem object describing a three-body problem IVP with the prescribed
        configuration.

    References
    ----------
    .. [1] Hairer, E., Norsett, S. and Wanner, G..
        Solving Ordinary Differential Equations I.
        Springer Series in Computational Mathematics, 1993.
    """

    def rhs(t, y):
        mu = 0.012277471  # a constant (standardised moon mass)
        mp = 1 - mu
        D1 = ((y[0] + mu) ** 2 + y[1] ** 2) ** (3 / 2)
        D2 = ((y[0] - mp) ** 2 + y[1] ** 2) ** (3 / 2)
        y1p = y[0] + 2 * y[3] - mp * (y[0] + mu) / D1 - mu * (y[0] - mp) / D2
        y2p = y[1] - 2 * y[2] - mp * y[1] / D1 - mu * y[1] / D2
        return np.array([y[2], y[3], y1p, y2p])

    if y0 is None:
        y0 = np.array([0.994, 0, 0, -2.00158510637908252240537862224])

    return InitialValueProblem(f=rhs, t0=t0, tmax=tmax, y0=y0)


def vanderpol(t0=0.0, tmax=30, y0=None, params=1e1):
    r"""Initial value problem (IVP) based on the Van der Pol Oscillator, implemented in `jax`.

    This function implements the second-order Van-der-Pol Oscillator as a system
    of first-order ODEs.
    The Van der Pol Oscillator is defined as

    .. math::

        f(t, y) =
        \begin{pmatrix}
            y_2 \\
            \mu \cdot (1 - y_1^2)y_2 - y_1
        \end{pmatrix}

    for a constant parameter  :math:`\mu`.
    :math:`\mu` determines the stiffness of the problem, where
    the larger :math:`\mu` is chosen, the more stiff the problem becomes.
    Default is :math:`\mu = 0.1`.
    This implementation includes the Jacobian :math:`J_f` of :math:`f`.

    Parameters
    ----------
    t0 : float
        Initial time point. Leftmost point of the integration domain.
    tmax : float
        Final time point. Rightmost point of the integration domain.
    y0 : np.ndarray,
        *(shape=(2, ))* -- Initial value of the problem.
    params : (float), optional
        Parameter :math:`\mu` for the Van der Pol Equations
        Default is :math:`\mu=0.1`.

    Returns
    -------
    InitialValueProblem
        InitialValueProblem object describing the Van der Pol Oscillator IVP with the prescribed
        configuration.
    """

    if y0 is None:
        y0 = np.array([2.0, 0.0])

    def rhs(t, y, params=params):
        y1, y2 = y
        if isinstance(params, float):
            mu = params
        else:
            (mu,) = params

        return np.array([y2, mu * (1.0 - y1 ** 2) * y2 - y1])

    def jac(t, y, params=params):
        y1, y2 = y
        if isinstance(params, float):
            mu = params
        else:
            (mu,) = params

        return np.array([[0.0, 1.0], [-2.0 * mu * y2 * y1 - 1.0, mu * (1.0 - y1 ** 2)]])

    return InitialValueProblem(f=rhs, t0=t0, tmax=tmax, y0=y0, df=jac)


def rigidbody(t0=0.0, tmax=20.0, y0=None, params=(-2.0, 1.25, -0.5)):
    r"""Initial value problem (IVP) for rigid body dynamics without external forces

    The rigid body dynamics without external forces is defined through

    .. math::

        f(t, y) =
        \begin{pmatrix}
            y_2 y_3 \\
            -y_1 y_3 \\
            -0.51 \cdot y_1 y_2
        \end{pmatrix}

    The ODE system has no parameters.
    This implementation includes the Jacobian :math:`J_f` of :math:`f`.

    Parameters
    ----------
    t0
        Initial time. Default is 0.0
    tmax
        Final time. Default is 20.0
    y0
        *(shape=(3, ))* -- Initial value. Default is ``[]1., 0., 0.9]``.
    params
        Parameter of the rigid body problem. Default is ``(-2.0, 1.25, -0.5)``.

    Returns
    -------
    InitialValueProblem
        InitialValueProblem object describing the rigid body dynamics IVP with the prescribed
        configuration.
    """
    if y0 is None:
        y0 = np.array([1.0, 0.0, 0.9])

    def rhs(t, y, params=params):
        p1, p2, p3 = params
        y1, y2, y3 = y
        return np.array([p1 * y2 * y3, p2 * y1 * y3, p3 * y1 * y2])

    def jac(t, y, params=params):
        p1, p2, p3 = params

        y1, y2, y3 = y
        return np.array(
            [[0.0, p1 * y3, p1 * y2], [p2 * y3, 0.0, p2 * y1], [p3 * y2, p3 * y1, 0.0]]
        )

    return InitialValueProblem(f=rhs, t0=t0, tmax=tmax, y0=y0, df=jac)



"""Particle filtering posterior."""

from typing import Optional

import numpy as np

from probnum import randvars
from probnum.filtsmooth.timeseriesposterior import (
    DenseOutputLocationArgType,
    TimeSeriesPosterior,
)
from probnum.type import FloatArgType, RandomStateArgType, ShapeArgType


class ParticleFilterPosterior(TimeSeriesPosterior):
    """Posterior distribution of a particle filter.."""

    def __call__(self, t):
        raise NotImplementedError("Particle filters do not provide dense output.")

    # The methods below are not implemented (yet?).

    def interpolate(self, t: FloatArgType) -> randvars.RandomVariable:
        raise NotImplementedError

    def sample(
        self,
        t: Optional[DenseOutputLocationArgType] = None,
        size: Optional[ShapeArgType] = (),
        random_state: Optional[RandomStateArgType] = None,
    ) -> np.ndarray:
        raise NotImplementedError("Sampling is not implemented.")

    def transform_base_measure_realizations(
        self,
        base_measure_realizations: np.ndarray,
        t: Optional[DenseOutputLocationArgType] = None,
    ) -> np.ndarray:
        raise NotImplementedError(
            "Transforming base measure realizations is not implemented."
        )
